@@ -13,24 +13,35 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
 
-from ctypes import Structure, c_char_p, c_uint, c_int, \
-		c_char, c_void_p, c_bool, create_string_buffer, \
-		POINTER as _POINTER, cdll as _cdll, memmove as _memmove, byref as _byref
+from ctypes import Structure, c_char_p, c_uint, c_int, c_size_t, \
+		c_ssize_t, c_char, c_void_p, c_bool, create_string_buffer, \
+		POINTER as _POINTER, CDLL as _cdll, memmove as _memmove, byref as _byref
 from os import strerror as _strerror
 from platform import system as _system
+import weakref
+
+if 'Windows' in _system():
+	from ctypes import get_last_error
+else:
+	from ctypes import get_errno
 
 def _checkNull(result, func, arguments):
 	if result:
 		return result
 	else:
-		raise Exception("Null pointer")
+		err = get_last_error() if 'Windows' in _system() else get_errno()
+		raise OSError(err, _strerror(err))
 
 def _checkNegative(result, func, arguments):
 	if result >= 0:
 		return result
 	else:
-		raise Exception("Error: " + _strerror(-result))
+		raise OSError(-result, _strerror(-result))
 
+class _ScanContext(Structure):
+	pass
+class _ContextInfo(Structure):
+	pass
 class _Context(Structure):
 	pass
 class _Device(Structure):
@@ -40,12 +51,47 @@ class _Channel(Structure):
 class _Buffer(Structure):
 	pass
 
+_ScanContextPtr = _POINTER(_ScanContext)
+_ContextInfoPtr = _POINTER(_ContextInfo)
 _ContextPtr = _POINTER(_Context)
 _DevicePtr = _POINTER(_Device)
 _ChannelPtr = _POINTER(_Channel)
 _BufferPtr = _POINTER(_Buffer)
 
-_lib = _cdll.LoadLibrary('libiio.dll' if 'Windows' in _system() else 'libiio.so.0')
+_lib = _cdll('libiio.dll' if 'Windows' in _system() else 'libiio.so.0',
+		use_errno = True, use_last_error = True)
+
+_get_backends_count = _lib.iio_get_backends_count
+_get_backends_count.restype = c_uint
+
+_get_backend = _lib.iio_get_backend
+_get_backend.argtypes = (c_uint, )
+_get_backend.restype = c_char_p
+_get_backend.errcheck = _checkNull
+
+_create_scan_context = _lib.iio_create_scan_context
+_create_scan_context.argtypes = (c_char_p, c_uint)
+_create_scan_context.restype = _ScanContextPtr
+_create_scan_context.errcheck = _checkNull
+
+_destroy_scan_context = _lib.iio_scan_context_destroy
+_destroy_scan_context.argtypes = (_ScanContextPtr, )
+
+_get_context_info_list = _lib.iio_scan_context_get_info_list
+_get_context_info_list.argtypes = (_ScanContextPtr, _POINTER(_POINTER(_ContextInfoPtr)))
+_get_context_info_list.restype = c_ssize_t
+_get_context_info_list.errcheck = _checkNegative
+
+_context_info_list_free = _lib.iio_context_info_list_free
+_context_info_list_free.argtypes = (_POINTER(_ContextInfoPtr), )
+
+_context_info_get_description = _lib.iio_context_info_get_description
+_context_info_get_description.argtypes = (_ContextInfoPtr, )
+_context_info_get_description.restype = c_char_p
+
+_context_info_get_uri = _lib.iio_context_info_get_uri
+_context_info_get_uri.argtypes = (_ContextInfoPtr, )
+_context_info_get_uri.restype = c_char_p
 
 _new_local = _lib.iio_create_local_context
 _new_local.restype = _ContextPtr
@@ -93,6 +139,15 @@ _get_version.restype = c_int
 _get_version.argtypes = (_ContextPtr, _POINTER(c_uint), _POINTER(c_uint), c_char_p, )
 _get_version.errcheck = _checkNegative
 
+_get_attrs_count = _lib.iio_context_get_attrs_count
+_get_attrs_count.restype = c_uint
+_get_attrs_count.argtypes = (_ContextPtr, )
+
+_get_attr = _lib.iio_context_get_attr
+_get_attr.restype = c_int
+_get_attr.argtypes = (_ContextPtr, c_uint, _POINTER(c_char_p), _POINTER(c_char_p))
+_get_attr.errcheck = _checkNegative
+
 _devices_count = _lib.iio_context_get_devices_count
 _devices_count.restype = c_uint
 _devices_count.argtypes = (_ContextPtr, )
@@ -131,12 +186,12 @@ _d_get_attr.argtypes = (_DevicePtr, )
 _d_get_attr.errcheck = _checkNull
 
 _d_read_attr = _lib.iio_device_attr_read
-_d_read_attr.restype = c_int
-_d_read_attr.argtypes = (_DevicePtr, c_char_p, c_char_p, c_uint)
+_d_read_attr.restype = c_ssize_t
+_d_read_attr.argtypes = (_DevicePtr, c_char_p, c_char_p, c_size_t)
 _d_read_attr.errcheck = _checkNegative
 
 _d_write_attr = _lib.iio_device_attr_write
-_d_write_attr.restype = c_int
+_d_write_attr.restype = c_ssize_t
 _d_write_attr.argtypes = (_DevicePtr, c_char_p, c_char_p)
 _d_write_attr.errcheck = _checkNegative
 
@@ -150,12 +205,12 @@ _d_get_debug_attr.argtypes = (_DevicePtr, )
 _d_get_debug_attr.errcheck = _checkNull
 
 _d_read_debug_attr = _lib.iio_device_debug_attr_read
-_d_read_debug_attr.restype = c_int
-_d_read_debug_attr.argtypes = (_DevicePtr, c_char_p, c_char_p, c_uint)
+_d_read_debug_attr.restype = c_ssize_t
+_d_read_debug_attr.argtypes = (_DevicePtr, c_char_p, c_char_p, c_size_t)
 _d_read_debug_attr.errcheck = _checkNegative
 
 _d_write_debug_attr = _lib.iio_device_debug_attr_write
-_d_write_debug_attr.restype = c_int
+_d_write_debug_attr.restype = c_ssize_t
 _d_write_debug_attr.argtypes = (_DevicePtr, c_char_p, c_char_p)
 _d_write_debug_attr.errcheck = _checkNegative
 
@@ -234,12 +289,12 @@ _c_get_filename.argtypes = (_ChannelPtr, c_char_p, )
 _c_get_filename.errcheck = _checkNull
 
 _c_read_attr = _lib.iio_channel_attr_read
-_c_read_attr.restype = c_int
-_c_read_attr.argtypes = (_ChannelPtr, c_char_p, c_char_p, c_uint)
+_c_read_attr.restype = c_ssize_t
+_c_read_attr.argtypes = (_ChannelPtr, c_char_p, c_char_p, c_size_t)
 _c_read_attr.errcheck = _checkNegative
 
 _c_write_attr = _lib.iio_channel_attr_write
-_c_write_attr.restype = c_int
+_c_write_attr.restype = c_ssize_t
 _c_write_attr.argtypes = (_ChannelPtr, c_char_p, c_char_p)
 _c_write_attr.errcheck = _checkNegative
 
@@ -254,36 +309,36 @@ _c_is_enabled.restype = c_bool
 _c_is_enabled.argtypes = (_ChannelPtr, )
 
 _c_read = _lib.iio_channel_read
-_c_read.restype = c_uint
-_c_read.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_uint, )
+_c_read.restype = c_ssize_t
+_c_read.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_size_t, )
 
 _c_read_raw = _lib.iio_channel_read_raw
-_c_read_raw.restype = c_uint
-_c_read_raw.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_uint, )
+_c_read_raw.restype = c_ssize_t
+_c_read_raw.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_size_t, )
 
 _c_write = _lib.iio_channel_write
-_c_write.restype = c_uint
-_c_write.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_uint, )
+_c_write.restype = c_ssize_t
+_c_write.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_size_t, )
 
 _c_write_raw = _lib.iio_channel_write_raw
-_c_write_raw.restype = c_uint
-_c_write_raw.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_uint, )
+_c_write_raw.restype = c_ssize_t
+_c_write_raw.argtypes = (_ChannelPtr, _BufferPtr, c_void_p, c_size_t, )
 
 _create_buffer = _lib.iio_device_create_buffer
 _create_buffer.restype = _BufferPtr
-_create_buffer.argtypes = (_DevicePtr, c_uint, c_bool, )
+_create_buffer.argtypes = (_DevicePtr, c_size_t, c_bool, )
 _create_buffer.errcheck = _checkNull
 
 _buffer_destroy = _lib.iio_buffer_destroy
 _buffer_destroy.argtypes = (_BufferPtr, )
 
 _buffer_refill = _lib.iio_buffer_refill
-_buffer_refill.restype = c_int
+_buffer_refill.restype = c_ssize_t
 _buffer_refill.argtypes = (_BufferPtr, )
 _buffer_refill.errcheck = _checkNegative
 
 _buffer_push_partial = _lib.iio_buffer_push_partial
-_buffer_push_partial.restype = c_int
+_buffer_push_partial.restype = c_ssize_t
 _buffer_push_partial.argtypes = (_BufferPtr, c_uint, )
 _buffer_push_partial.errcheck = _checkNegative
 
@@ -300,13 +355,15 @@ def _get_lib_version():
 	minor = c_uint()
 	buf = create_string_buffer(8)
 	_get_library_version(_byref(major), _byref(minor), buf)
-	return (major.value, minor.value, buf.value )
+	return (major.value, minor.value, buf.value.decode('ascii') )
 
 version = _get_lib_version()
+backends = [ _get_backend(x).decode('ascii') for x in range(0, _get_backends_count()) ]
 
 class _Attr(object):
 	def __init__(self, name, filename = None):
 		self._name = name
+		self._name_ascii = name.encode('ascii')
 		self._filename = name if filename is None else filename
 
 	def __str__(self):
@@ -323,16 +380,16 @@ class ChannelAttr(_Attr):
 	"""Represents an attribute of a channel."""
 
 	def __init__(self, channel, name):
-		super(ChannelAttr, self).__init__(name, _c_get_filename(channel, name))
+		super(ChannelAttr, self).__init__(name, _c_get_filename(channel, name.encode('ascii')).decode('ascii'))
 		self._channel = channel
 
 	def _Attr__read(self):
 		buf = create_string_buffer(1024)
-		_c_read_attr(self._channel, self.name, buf, len(buf))
-		return buf.value
+		_c_read_attr(self._channel, self._name_ascii, buf, len(buf))
+		return buf.value.decode('ascii')
 
 	def _Attr__write(self, value):
-		_c_write_attr(self._channel, self.name, value)
+		_c_write_attr(self._channel, self._name_ascii, value.encode('ascii'))
 
 class DeviceAttr(_Attr):
 	"""Represents an attribute of an IIO device."""
@@ -343,11 +400,11 @@ class DeviceAttr(_Attr):
 
 	def _Attr__read(self):
 		buf = create_string_buffer(1024)
-		_d_read_attr(self._device, self.name, buf, len(buf))
-		return buf.value
+		_d_read_attr(self._device, self._name_ascii, buf, len(buf))
+		return buf.value.decode('ascii')
 
 	def _Attr__write(self, value):
-		_d_write_attr(self._device, self.name, value)
+		_d_write_attr(self._device, self._name_ascii, value.encode('ascii'))
 
 class DeviceDebugAttr(DeviceAttr):
 	"""Represents a debug attribute of an IIO device."""
@@ -357,20 +414,21 @@ class DeviceDebugAttr(DeviceAttr):
 
 	def _Attr__read(self):
 		buf = create_string_buffer(1024)
-		_d_read_debug_attr(self._device, self.name, buf, len(buf))
-		return buf.value
+		_d_read_debug_attr(self._device, self._name_ascii, buf, len(buf))
+		return buf.value.decode('ascii')
 
 	def _Attr__write(self, value):
-		_d_write_debug_attr(self._device, self.name, value)
+		_d_write_debug_attr(self._device, self._name_ascii, value.encode('ascii'))
 
 class Channel(object):
-	def __init__(self, dev, _channel):
-		self.dev = dev
+	def __init__(self, _channel):
 		self._channel = _channel
 		self._attrs = { name : ChannelAttr(_channel, name) for name in \
-				[_c_get_attr(_channel, x) for x in xrange(0, _c_attr_count(_channel))] }
-		self._id = _c_get_id(self._channel)
-		self._name = _c_get_name(self._channel)
+				[_c_get_attr(_channel, x).decode('ascii') for x in range(0, _c_attr_count(_channel))] }
+		self._id = _c_get_id(self._channel).decode('ascii')
+
+		name_raw = _c_get_name(self._channel)
+		self._name = name_raw.decode('ascii') if name_raw is not None else None
 		self._output = _c_is_output(self._channel)
 		self._scan_element = _c_is_scan_element(self._channel)
 
@@ -453,7 +511,6 @@ class Buffer(object):
 		returns: type=iio.Buffer
 			An new instance of this class
 		"""
-		self.dev = device
 		try:
 			self._buffer = _create_buffer(device._device, samples_count, cyclic)
 		except:
@@ -462,6 +519,10 @@ class Buffer(object):
 		self._length = samples_count * device.sample_size
 		self._samples_count = samples_count
 
+		self._ctx = device.ctx() 
+		# Holds a reference to the corresponding IIO Context. This ensures that
+		# every iio.Buffer object is destroyed before its corresponding IIO Context.
+
 	def __del__(self):
 		if self._buffer is not None:
 			_buffer_destroy(self._buffer)
@@ -522,20 +583,21 @@ class Buffer(object):
 		return length
 
 class _DeviceOrTrigger(object):
-	def __init__(self, ctx, _device):
-		self.ctx = ctx
+	def __init__(self, _device):
 		self._device = _device
 		self._attrs = { name : DeviceAttr(_device, name) for name in \
-				[_d_get_attr(_device, x) for x in xrange(0, _d_attr_count(_device))] }
+				[_d_get_attr(_device, x).decode('ascii') for x in range(0, _d_attr_count(_device))] }
 		self._debug_attrs = { name: DeviceDebugAttr(_device, name) for name in \
-				[_d_get_debug_attr(_device, x) for x in xrange(0, _d_debug_attr_count(_device))] }
+				[_d_get_debug_attr(_device, x).decode('ascii') for x in range(0, _d_debug_attr_count(_device))] }
 
 		# TODO(pcercuei): Use a dictionary for the channels.
-		chans = [ Channel(self, _get_channel(self._device, x)) 
-			for x in xrange(0, _channels_count(self._device)) ]
+		chans = [ Channel(_get_channel(self._device, x))
+			for x in range(0, _channels_count(self._device)) ]
 		self._channels = sorted(chans, key=lambda c: c.id)
-		self._id = _d_get_id(self._device)
-		self._name = _d_get_name(self._device)
+		self._id = _d_get_id(self._device).decode('ascii')
+
+		name_raw = _d_get_name(self._device)
+		self._name = name_raw.decode('ascii') if name_raw is not None else None
 
 	def reg_write(self, reg, value):
 		"""
@@ -578,8 +640,9 @@ class _DeviceOrTrigger(object):
 		returns: type=iio.Device or type=iio.Trigger
 			The IIO Device
 		"""
-		return (filter(lambda x: (name_or_id == x.name or name_or_id == x.id) \
-				and x.output == is_output, self.channels) or [None])[0]
+		return next((x for x in self.channels \
+				if name_or_id == x.name or name_or_id == x.id and \
+				x.output == is_output), None)
 
 	def set_kernel_buffers_count(self, count):
 		"""
@@ -616,8 +679,8 @@ class _DeviceOrTrigger(object):
 class Trigger(_DeviceOrTrigger):
 	"""Contains the representation of an IIO device that can act as a trigger."""
 
-	def __init__(self, ctx, _device):
-		super(Trigger, self).__init__(ctx, _device)
+	def __init__(self, _device):
+		super(Trigger, self).__init__(_device)
 
 	def _get_rate(self):
 		return int(self._attrs['frequency'].value)
@@ -632,7 +695,8 @@ class Device(_DeviceOrTrigger):
 	"""Contains the representation of an IIO device."""
 
 	def __init__(self, ctx, _device):
-		super(Device, self).__init__(ctx, _device)
+		super(Device, self).__init__(_device)
+		self.ctx = weakref.ref(ctx)
 
 	def _set_trigger(self, trigger):
 		_d_set_trigger(self._device, trigger._device if trigger else None)
@@ -641,7 +705,7 @@ class Device(_DeviceOrTrigger):
 		value = _Device()
 		_d_get_trigger(self._device, _byref(value))
 
-		for dev in self.ctx._devices:
+		for dev in self.ctx()._devices:
 			if value == dev._device:
 				return dev
 		return None
@@ -664,25 +728,34 @@ class Context(object):
 		If set to an empty string, the server will be discovered using ZeroConf.
 		If the environment variable is not set, a local context will be created instead.
 		"""
+		self._context = None
+
 		if(_context is None):
 			self._context = _new_default()
-		elif type(_context) is str:
-			self._context = _new_uri(_context)
+		elif type(_context) is str or type(_context) is unicode:
+			self._context = _new_uri(_context.encode('ascii'))
 		else:
 			self._context = _context
 
+		self._attrs = {}
+		for x in range(0, _get_attrs_count(self._context)):
+			str1 = c_char_p()
+			str2 = c_char_p()
+			_get_attr(self._context, x, _byref(str1), _byref(str2))
+			self._attrs[str1.value.decode('ascii')] = str2.value.decode('ascii')
+
 		# TODO(pcercuei): Use a dictionary for the devices.
-		self._devices = [ Trigger(self, dev) if _d_is_trigger(dev) else Device(self, dev) for dev in \
-				[ _get_device(self._context, x) for x in xrange(0, _devices_count(self._context)) ]]
-		self._name = _get_name(self._context)
-		self._description = _get_description(self._context)
-		self._xml = _get_xml(self._context)
+		self._devices = [ Trigger(dev) if _d_is_trigger(dev) else Device(self, dev) for dev in \
+				[ _get_device(self._context, x) for x in range(0, _devices_count(self._context)) ]]
+		self._name = _get_name(self._context).decode('ascii')
+		self._description = _get_description(self._context).decode('ascii')
+		self._xml = _get_xml(self._context).decode('ascii')
 
 		major = c_uint()
 		minor = c_uint()
 		buf = create_string_buffer(8)
 		_get_version(self._context, _byref(major), _byref(minor), buf)
-		self._version = (major.value, minor.value, buf.value )
+		self._version = (major.value, minor.value, buf.value.decode('ascii') )
 
 	def __del__(self):
 		if(self._context is not None):
@@ -719,7 +792,8 @@ class Context(object):
 		returns: type=iio.Device or type=iio.Trigger
 			The IIO Device
 		"""
-		return (filter(lambda x: name_or_id == x.name or name_or_id == x.id, self.devices) or [None])[0]
+		return next((x for x in self.devices \
+				if name_or_id == x.name or name_or_id == x.id), None)
 
 	name = property(lambda self: self._name, None, None, \
 			"Name of this IIO context.\n\ttype=str")
@@ -729,6 +803,8 @@ class Context(object):
 			"XML representation of the current context.\n\ttype=str")
 	version = property(lambda self: self._version, None, None, \
 			"Version of the backend.\n\ttype=(int, int, str)")
+	attrs = property(lambda self: self._attrs, None, None, \
+			"List of context-specific attributes\n\ttype=dict of str objects")
 	devices = property(lambda self: self._devices, None, None, \
 			"List of devices contained in this context.\n\ttype=list of iio.Device and iio.Trigger objects")
 
@@ -755,7 +831,7 @@ class XMLContext(Context):
 		returns: type=iio.XMLContext
 			An new instance of this class
 		"""
-		ctx = _new_xml(xmlfile)
+		ctx = _new_xml(xmlfile.encode('ascii'))
 		super(XMLContext, self).__init__(ctx)
 
 class NetworkContext(Context):
@@ -770,5 +846,19 @@ class NetworkContext(Context):
 		returns: type=iio.NetworkContext
 			An new instance of this class
 		"""
-		ctx = _new_network(hostname)
+		ctx = _new_network(hostname.encode('ascii') if hostname is not None else None)
 		super(NetworkContext, self).__init__(ctx)
+
+def scan_contexts():
+	d = dict()
+	ptr = _POINTER(_ContextInfoPtr)()
+
+	ctx = _create_scan_context(None, 0)
+	nb = _get_context_info_list(ctx, _byref(ptr));
+
+	for i in range(0, nb):
+		d[_context_info_get_uri(ptr[i]).decode('ascii')] = _context_info_get_description(ptr[i]).decode('ascii')
+
+	_context_info_list_free(ptr)
+	_destroy_scan_context(ctx)
+	return d
