@@ -51,6 +51,11 @@ static const char * const iio_chan_type_name_spec[] = {
 	[IIO_CONCENTRATION] = "concentration",
 	[IIO_RESISTANCE] = "resistance",
 	[IIO_PH] = "ph",
+	[IIO_UVINDEX] = "uvindex",
+	[IIO_ELECTRICALCONDUCTIVITY] = "electricalconductivity",
+	[IIO_COUNT] = "count",
+	[IIO_INDEX] = "index",
+	[IIO_GRAVITY] = "gravity",
 };
 
 static const char * const modifier_names[] = {
@@ -73,6 +78,7 @@ static const char * const modifier_names[] = {
 	[IIO_MOD_LIGHT_RED] = "red",
 	[IIO_MOD_LIGHT_GREEN] = "green",
 	[IIO_MOD_LIGHT_BLUE] = "blue",
+	[IIO_MOD_LIGHT_UV] = "uv",
 	[IIO_MOD_QUATERNION] = "quaternion",
 	[IIO_MOD_TEMP_AMBIENT] = "ambient",
 	[IIO_MOD_TEMP_OBJECT] = "object",
@@ -135,10 +141,10 @@ void iio_channel_init_finalize(struct iio_channel *chn)
 		if (strncmp(iio_chan_type_name_spec[i], chn->id, len) != 0)
 			continue;
 		/* Type must be followed by either a '_' or a digit */
-		if (chn->id[len] != '_' && chn->id[len] < '0' && chn->id[len] > '9')
+		if (chn->id[len] != '_' && (chn->id[len] < '0' || chn->id[len] > '9'))
 			continue;
 
-		chn->type = i;
+		chn->type = (enum iio_chan_type) i;
 	}
 
 	mod = strchr(chn->id, '_');
@@ -153,11 +159,8 @@ void iio_channel_init_finalize(struct iio_channel *chn)
 		len = strlen(modifier_names[i]);
 		if (strncmp(modifier_names[i], mod, len) != 0)
 			continue;
-		/* Modifier must be followed by a '_' */
-		if (mod[len] != '_')
-			continue;
 
-		chn->modifier = i;
+		chn->modifier = (enum iio_modifier) i;
 		break;
 	}
 }
@@ -175,32 +178,35 @@ static char *get_attr_xml(struct iio_channel_attr *attr, size_t *length)
 
 	*length = len - 1; /* Skip the \0 */
 	if (attr->filename)
-		snprintf(str, len, "<attribute name=\"%s\" filename=\"%s\" />",
+		iio_snprintf(str, len, "<attribute name=\"%s\" filename=\"%s\" />",
 				attr->name, attr->filename);
 	else
-		snprintf(str, len, "<attribute name=\"%s\" />", attr->name);
+		iio_snprintf(str, len, "<attribute name=\"%s\" />", attr->name);
 	return str;
 }
 
 static char * get_scan_element(const struct iio_channel *chn, size_t *length)
 {
-	char buf[1024], *str;
+	char buf[1024], repeat[12] = "", *str;
 	char processed = (chn->format.is_fully_defined ? 'A' - 'a' : 0);
 
-	snprintf(buf, sizeof(buf), "<scan-element index=\"%li\" "
-			"format=\"%ce:%c%u/%u&gt;&gt;%u\" />",
+	if (chn->format.repeat > 1)
+		iio_snprintf(repeat, sizeof(repeat), "X%u", chn->format.repeat);
+
+	iio_snprintf(buf, sizeof(buf), "<scan-element index=\"%li\" "
+			"format=\"%ce:%c%u/%u%s&gt;&gt;%u\" />",
 			chn->index, chn->format.is_be ? 'b' : 'l',
 			chn->format.is_signed ? 's' + processed : 'u' + processed,
-			chn->format.bits, chn->format.length,
+			chn->format.bits, chn->format.length, repeat,
 			chn->format.shift);
 
 	if (chn->format.with_scale) {
 		char *ptr = strrchr(buf, '\0');
-		snprintf(ptr - 2, buf + sizeof(buf) - ptr + 2,
+		iio_snprintf(ptr - 2, buf + sizeof(buf) - ptr + 2,
 				"scale=\"%f\" />", chn->format.scale);
 	}
 
-	str = _strdup(buf);
+	str = iio_strdup(buf);
 	if (str)
 		*length = strlen(str);
 	return str;
@@ -244,7 +250,7 @@ char * iio_channel_get_xml(const struct iio_channel *chn, size_t *length)
 	if (!str)
 		goto err_free_attrs;
 
-	snprintf(str, len, "<channel id=\"%s\"", chn->id);
+	iio_snprintf(str, len, "<channel id=\"%s\"", chn->id);
 	ptr = strrchr(str, '\0');
 
 	if (chn->name) {
@@ -392,19 +398,19 @@ const struct iio_data_format * iio_channel_get_data_format(
 bool iio_channel_is_enabled(const struct iio_channel *chn)
 {
 	return chn->index >= 0 && chn->dev->mask &&
-		TEST_BIT(chn->dev->mask, chn->index);
+		TEST_BIT(chn->dev->mask, chn->number);
 }
 
 void iio_channel_enable(struct iio_channel *chn)
 {
 	if (chn->is_scan_element && chn->index >= 0 && chn->dev->mask)
-		SET_BIT(chn->dev->mask, chn->index);
+		SET_BIT(chn->dev->mask, chn->number);
 }
 
 void iio_channel_disable(struct iio_channel *chn)
 {
 	if (chn->index >= 0 && chn->dev->mask)
-		CLEAR_BIT(chn->dev->mask, chn->index);
+		CLEAR_BIT(chn->dev->mask, chn->number);
 }
 
 void free_channel(struct iio_channel *chn)
@@ -519,33 +525,46 @@ static void mask_upper_bits(uint8_t *dst, size_t bits, size_t len)
 void iio_channel_convert(const struct iio_channel *chn,
 		void *dst, const void *src)
 {
+	uintptr_t src_ptr = (uintptr_t) src, dst_ptr = (uintptr_t) dst;
 	unsigned int len = chn->format.length / 8;
+	ptrdiff_t end = len * chn->format.repeat;
+	uintptr_t end_ptr = src_ptr + end;
 #if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
 	bool swap = chn->format.is_be;
 #else
 	bool swap = !chn->format.is_be;
 #endif
 
-	if (len == 1 || !swap)
-		memcpy(dst, src, len);
-	else
-		byte_swap(dst, src, len);
-
-	if (chn->format.shift)
-		shift_bits(dst, chn->format.shift, len, false);
-
-	if (!chn->format.is_fully_defined) {
-		if (chn->format.is_signed)
-			sign_extend(dst, chn->format.bits, len);
+	for (src_ptr = (uintptr_t) src; src_ptr < end_ptr;
+			src_ptr += len, dst_ptr += len) {
+		if (len == 1 || !swap)
+			memcpy((void *) dst_ptr, (const void *) src_ptr, len);
 		else
-			mask_upper_bits(dst, chn->format.bits, len);
+			byte_swap((void *) dst_ptr, (const void *) src_ptr,
+				len);
+
+		if (chn->format.shift)
+			shift_bits((void *) dst_ptr, chn->format.shift, len,
+				false);
+
+		if (!chn->format.is_fully_defined) {
+			if (chn->format.is_signed)
+				sign_extend((void *) dst_ptr,
+					chn->format.bits, len);
+			else
+				mask_upper_bits((void *) dst_ptr,
+					chn->format.bits, len);
+		}
 	}
 }
 
 void iio_channel_convert_inverse(const struct iio_channel *chn,
 		void *dst, const void *src)
 {
+	uintptr_t src_ptr = (uintptr_t) src, dst_ptr = (uintptr_t) dst;
 	unsigned int len = chn->format.length / 8;
+	ptrdiff_t end = len * chn->format.repeat;
+	uintptr_t end_ptr = dst_ptr + end;
 #if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
 	bool swap = chn->format.is_be;
 #else
@@ -557,23 +576,26 @@ void iio_channel_convert_inverse(const struct iio_channel *chn,
 	if (len > sizeof(buf))
 		return;
 
-	memcpy(buf, src, len);
-	mask_upper_bits(buf, chn->format.bits, len);
+	for (dst_ptr = (uintptr_t) dst; dst_ptr < end_ptr;
+			src_ptr += len, dst_ptr += len) {
+		memcpy(buf, (const void *) src_ptr, len);
+		mask_upper_bits(buf, chn->format.bits, len);
 
-	if (chn->format.shift)
-		shift_bits(buf, chn->format.shift, len, true);
+		if (chn->format.shift)
+			shift_bits(buf, chn->format.shift, len, true);
 
-	if (len == 1 || !swap)
-		memcpy(dst, buf, len);
-	else
-		byte_swap(dst, buf, len);
+		if (len == 1 || !swap)
+			memcpy((void *) dst_ptr, buf, len);
+		else
+			byte_swap((void *) dst_ptr, buf, len);
+	}
 }
 
 size_t iio_channel_read_raw(const struct iio_channel *chn,
 		struct iio_buffer *buf, void *dst, size_t len)
 {
 	uintptr_t src_ptr, dst_ptr = (uintptr_t) dst, end = dst_ptr + len;
-	unsigned int length = chn->format.length / 8;
+	unsigned int length = chn->format.length / 8 * chn->format.repeat;
 	uintptr_t buf_end = (uintptr_t) iio_buffer_end(buf);
 	ptrdiff_t buf_step = iio_buffer_step(buf);
 
@@ -588,7 +610,7 @@ size_t iio_channel_read(const struct iio_channel *chn,
 		struct iio_buffer *buf, void *dst, size_t len)
 {
 	uintptr_t src_ptr, dst_ptr = (uintptr_t) dst, end = dst_ptr + len;
-	unsigned int length = chn->format.length / 8;
+	unsigned int length = chn->format.length / 8 * chn->format.repeat;
 	uintptr_t buf_end = (uintptr_t) iio_buffer_end(buf);
 	ptrdiff_t buf_step = iio_buffer_step(buf);
 
@@ -604,7 +626,7 @@ size_t iio_channel_write_raw(const struct iio_channel *chn,
 		struct iio_buffer *buf, const void *src, size_t len)
 {
 	uintptr_t dst_ptr, src_ptr = (uintptr_t) src, end = src_ptr + len;
-	unsigned int length = chn->format.length / 8;
+	unsigned int length = chn->format.length / 8 * chn->format.repeat;
 	uintptr_t buf_end = (uintptr_t) iio_buffer_end(buf);
 	ptrdiff_t buf_step = iio_buffer_step(buf);
 
@@ -619,7 +641,7 @@ size_t iio_channel_write(const struct iio_channel *chn,
 		struct iio_buffer *buf, const void *src, size_t len)
 {
 	uintptr_t dst_ptr, src_ptr = (uintptr_t) src, end = src_ptr + len;
-	unsigned int length = chn->format.length / 8;
+	unsigned int length = chn->format.length / 8 * chn->format.repeat;
 	uintptr_t buf_end = (uintptr_t) iio_buffer_end(buf);
 	ptrdiff_t buf_step = iio_buffer_step(buf);
 
@@ -675,7 +697,7 @@ int iio_channel_attr_write_longlong(const struct iio_channel *chn,
 {
 	ssize_t ret;
 	char buf[1024];
-	snprintf(buf, sizeof(buf), "%lld", val);
+	iio_snprintf(buf, sizeof(buf), "%lld", val);
 	ret = iio_channel_attr_write(chn, attr, buf);
 	return ret < 0 ? ret : 0;
 }
@@ -719,7 +741,7 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	int ret;
+	int ret, buf_size;
 	char *buf, *ptr;
 	unsigned int i;
 
@@ -733,12 +755,26 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 		goto err_free_buf;
 
 	ptr = buf;
+	buf_size = ret;
 
 	for (i = 0; i < iio_channel_get_attrs_count(chn); i++) {
 		const char *attr = iio_channel_get_attr(chn, i);
-		int32_t len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
+		int32_t len;
 
+		if (buf_size < 4) {
+			ret = -EPROTO;
+			break;
+		}
+
+		len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
 		ptr += 4;
+		buf_size -= 4;
+
+		if (len > 0 && buf_size < len) {
+			ret = -EPROTO;
+			break;
+		}
+
 		if (len > 0) {
 			ret = cb(chn, attr, ptr, (size_t) len, data);
 			if (ret < 0)
@@ -747,6 +783,10 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 			if (len & 0x3)
 				len = ((len >> 2) + 1) << 2;
 			ptr += len;
+			if (len >= buf_size)
+				buf_size = 0;
+			else
+				buf_size -= len;
 		}
 	}
 
