@@ -24,19 +24,41 @@
 #include <stdio.h>
 #include <string.h>
 
-static char *get_attr_xml(const char *attr, size_t *length, bool is_debug)
+static char *get_attr_xml(const char *attr, size_t *length, enum iio_attr_type type)
 {
-	size_t len = sizeof("<attribute name=\"\" />") + strlen(attr)
-		+ (!is_debug ? 0 : sizeof("debug-") - 1);
-	char *str = malloc(len);
+	size_t len = sizeof("<attribute name=\"\" />") + strlen(attr);
+	char *str;
+
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			len += (sizeof("debug-") - 1);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			len += (sizeof("buffer-") - 1);
+			break;
+		default:
+			return NULL;
+	}
+
+	str = malloc(len);
 	if (!str)
 		return NULL;
 
 	*length = len - 1; /* Skip the \0 */
-	if (is_debug)
-		snprintf(str, len, "<debug-attribute name=\"%s\" />", attr);
-	else
-		snprintf(str, len, "<attribute name=\"%s\" />", attr);
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			iio_snprintf(str, len, "<attribute name=\"%s\" />", attr);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			iio_snprintf(str, len, "<debug-attribute name=\"%s\" />", attr);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			iio_snprintf(str, len, "<buffer-attribute name=\"%s\" />", attr);
+			break;
+	}
+
 	return str;
 }
 
@@ -45,8 +67,8 @@ char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
 {
 	size_t len = sizeof("<device id=\"\" name=\"\" ></device>")
 		+ strlen(dev->id) + (dev->name ? strlen(dev->name) : 0);
-	char *ptr, *str, **attrs, **channels, **debug_attrs;
-	size_t *attrs_len, *channels_len, *debug_attrs_len;
+	char *ptr, *str, **attrs, **channels, **buffer_attrs, **debug_attrs;
+	size_t *attrs_len, *channels_len, *buffer_attrs_len, *debug_attrs_len;
 	unsigned int i, j, k;
 
 	attrs_len = malloc(dev->nb_attrs * sizeof(*attrs_len));
@@ -58,7 +80,7 @@ char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
 		goto err_free_attrs_len;
 
 	for (i = 0; i < dev->nb_attrs; i++) {
-		char *xml = get_attr_xml(dev->attrs[i], &attrs_len[i], false);
+		char *xml = get_attr_xml(dev->attrs[i], &attrs_len[i], IIO_ATTR_TYPE_DEVICE);
 		if (!xml)
 			goto err_free_attrs;
 		attrs[i] = xml;
@@ -82,10 +104,28 @@ char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
 		len += channels_len[j];
 	}
 
+	buffer_attrs_len = malloc(dev->nb_buffer_attrs *
+			sizeof(*buffer_attrs_len));
+	if (!buffer_attrs_len)
+		goto err_free_channels;
+
+	buffer_attrs = malloc(dev->nb_buffer_attrs * sizeof(*buffer_attrs));
+	if (!buffer_attrs)
+		goto err_free_buffer_attrs_len;
+
+	for (k = 0; k < dev->nb_buffer_attrs; k++) {
+		char *xml = get_attr_xml(dev->buffer_attrs[k],
+				&buffer_attrs_len[k], IIO_ATTR_TYPE_BUFFER);
+		if (!xml)
+			goto err_free_buffer_attrs;
+		buffer_attrs[k] = xml;
+		len += buffer_attrs_len[k];
+	}
+
 	debug_attrs_len = malloc(dev->nb_debug_attrs *
 			sizeof(*debug_attrs_len));
 	if (!debug_attrs_len)
-		goto err_free_channels;
+		goto err_free_buffer_attrs;
 
 	debug_attrs = malloc(dev->nb_debug_attrs * sizeof(*debug_attrs));
 	if (!debug_attrs)
@@ -93,7 +133,7 @@ char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
 
 	for (k = 0; k < dev->nb_debug_attrs; k++) {
 		char *xml = get_attr_xml(dev->debug_attrs[k],
-				&debug_attrs_len[k], true);
+				&debug_attrs_len[k], IIO_ATTR_TYPE_DEBUG);
 		if (!xml)
 			goto err_free_debug_attrs;
 		debug_attrs[k] = xml;
@@ -104,7 +144,7 @@ char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
 	if (!str)
 		goto err_free_debug_attrs;
 
-	snprintf(str, len, "<device id=\"%s\"", dev->id);
+	iio_snprintf(str, len, "<device id=\"%s\"", dev->id);
 	ptr = strrchr(str, '\0');
 
 	if (dev->name) {
@@ -133,6 +173,15 @@ char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
 	free(attrs);
 	free(attrs_len);
 
+	for (i = 0; i < dev->nb_buffer_attrs; i++) {
+		strcpy(ptr, buffer_attrs[i]);
+		ptr += buffer_attrs_len[i];
+		free(buffer_attrs[i]);
+	}
+
+	free(buffer_attrs);
+	free(buffer_attrs_len);
+
 	for (i = 0; i < dev->nb_debug_attrs; i++) {
 		strcpy(ptr, debug_attrs[i]);
 		ptr += debug_attrs_len[i];
@@ -152,6 +201,12 @@ err_free_debug_attrs:
 	free(debug_attrs);
 err_free_debug_attrs_len:
 	free(debug_attrs_len);
+err_free_buffer_attrs:
+	while (k--)
+		free(buffer_attrs[k]);
+	free(buffer_attrs);
+err_free_buffer_attrs_len:
+	free(buffer_attrs_len);
 err_free_channels:
 	while (j--)
 		free(channels[j]);
@@ -233,6 +288,32 @@ const char * iio_device_find_attr(const struct iio_device *dev,
 	return NULL;
 }
 
+unsigned int iio_device_get_buffer_attrs_count(const struct iio_device *dev)
+{
+	return dev->nb_buffer_attrs;
+}
+
+const char * iio_device_get_buffer_attr(const struct iio_device *dev,
+		unsigned int index)
+{
+	if (index >= dev->nb_buffer_attrs)
+		return NULL;
+	else
+		return dev->buffer_attrs[index];
+}
+
+const char * iio_device_find_buffer_attr(const struct iio_device *dev,
+		const char *name)
+{
+	unsigned int i;
+	for (i = 0; i < dev->nb_buffer_attrs; i++) {
+		const char *attr = dev->buffer_attrs[i];
+		if (!strcmp(attr, name))
+			return attr;
+	}
+	return NULL;
+}
+
 const char * iio_device_find_debug_attr(const struct iio_device *dev,
 		const char *name)
 {
@@ -322,7 +403,7 @@ ssize_t iio_device_attr_read(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->read_device_attr)
 		return dev->ctx->ops->read_device_attr(dev,
-				attr, dst, len, false);
+				attr, dst, len, IIO_ATTR_TYPE_DEVICE);
 	else
 		return -ENOSYS;
 }
@@ -332,7 +413,7 @@ ssize_t iio_device_attr_write_raw(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->write_device_attr)
 		return dev->ctx->ops->write_device_attr(dev,
-				attr, src, len, false);
+				attr, src, len, IIO_ATTR_TYPE_DEVICE);
 	else
 		return -ENOSYS;
 }
@@ -343,6 +424,32 @@ ssize_t iio_device_attr_write(const struct iio_device *dev,
 	return iio_device_attr_write_raw(dev, attr, src, strlen(src) + 1);
 }
 
+ssize_t iio_device_buffer_attr_read(const struct iio_device *dev,
+		const char *attr, char *dst, size_t len)
+{
+	if (dev->ctx->ops->read_device_attr)
+		return dev->ctx->ops->read_device_attr(dev,
+				attr, dst, len, IIO_ATTR_TYPE_BUFFER);
+	else
+		return -ENOSYS;
+}
+
+ssize_t iio_device_buffer_attr_write_raw(const struct iio_device *dev,
+		const char *attr, const void *src, size_t len)
+{
+	if (dev->ctx->ops->write_device_attr)
+		return dev->ctx->ops->write_device_attr(dev,
+				attr, src, len, IIO_ATTR_TYPE_BUFFER);
+	else
+		return -ENOSYS;
+}
+
+ssize_t iio_device_buffer_attr_write(const struct iio_device *dev,
+		const char *attr, const char *src)
+{
+	return iio_device_buffer_attr_write_raw(dev, attr, src, strlen(src) + 1);
+}
+
 void iio_device_set_data(struct iio_device *dev, void *data)
 {
 	dev->userdata = data;
@@ -405,6 +512,10 @@ void free_device(struct iio_device *dev)
 		free(dev->attrs[i]);
 	if (dev->nb_attrs)
 		free(dev->attrs);
+	for (i = 0; i < dev->nb_buffer_attrs; i++)
+		free(dev->buffer_attrs[i]);
+	if (dev->nb_buffer_attrs)
+		free(dev->buffer_attrs);
 	for (i = 0; i < dev->nb_debug_attrs; i++)
 		free(dev->debug_attrs[i]);
 	if (dev->nb_debug_attrs)
@@ -427,23 +538,32 @@ ssize_t iio_device_get_sample_size_mask(const struct iio_device *dev,
 {
 	ssize_t size = 0;
 	unsigned int i;
+	const struct iio_channel *prev = NULL;
 
 	if (words != (dev->nb_channels + 31) / 32)
 		return -EINVAL;
 
 	for (i = 0; i < dev->nb_channels; i++) {
 		const struct iio_channel *chn = dev->channels[i];
-		unsigned int length = chn->format.length / 8;
+		unsigned int length = chn->format.length / 8 *
+			chn->format.repeat;
 
 		if (chn->index < 0)
 			break;
-		if (!TEST_BIT(mask, chn->index))
+		if (!TEST_BIT(mask, chn->number))
 			continue;
 
+		if (prev && chn->index == prev->index) {
+			prev = chn;
+			continue;
+		}
+
 		if (size % length)
 			size += 2 * length - (size % length);
 		else
 			size += length;
+
+		prev = chn;
 	}
 	return size;
 }
@@ -498,7 +618,7 @@ int iio_device_attr_write_longlong(const struct iio_device *dev,
 	ssize_t ret;
 	char buf[1024];
 
-	snprintf(buf, sizeof(buf), "%lld", val);
+	iio_snprintf(buf, sizeof(buf), "%lld", val);
 	ret = iio_device_attr_write(dev, attr, buf);
 
 	return ret < 0 ? ret : 0;
@@ -529,12 +649,88 @@ int iio_device_attr_write_bool(const struct iio_device *dev,
 	return ret < 0 ? ret : 0;
 }
 
+int iio_device_buffer_attr_read_longlong(const struct iio_device *dev,
+		const char *attr, long long *val)
+{
+	char *end, buf[1024];
+	long long value;
+	ssize_t ret = iio_device_buffer_attr_read(dev, attr, buf, sizeof(buf));
+	if (ret < 0)
+		return (int) ret;
+
+	value = strtoll(buf, &end, 0);
+	if (end == buf)
+		return -EINVAL;
+	*val = value;
+	return 0;
+}
+
+int iio_device_buffer_attr_read_bool(const struct iio_device *dev,
+		const char *attr, bool *val)
+{
+	long long value;
+	int ret = iio_device_buffer_attr_read_longlong(dev, attr, &value);
+	if (ret < 0)
+		return ret;
+
+	*val = !!value;
+	return 0;
+}
+
+int iio_device_buffer_attr_read_double(const struct iio_device *dev,
+		const char *attr, double *val)
+{
+	char buf[1024];
+	ssize_t ret = iio_device_buffer_attr_read(dev, attr, buf, sizeof(buf));
+	if (ret < 0)
+		return (int) ret;
+	else
+		return read_double(buf, val);
+}
+
+int iio_device_buffer_attr_write_longlong(const struct iio_device *dev,
+		const char *attr, long long val)
+{
+	ssize_t ret;
+	char buf[1024];
+
+	iio_snprintf(buf, sizeof(buf), "%lld", val);
+	ret = iio_device_buffer_attr_write(dev, attr, buf);
+
+	return ret < 0 ? ret : 0;
+}
+
+int iio_device_buffer_attr_write_double(const struct iio_device *dev,
+		const char *attr, double val)
+{
+	ssize_t ret;
+	char buf[1024];
+
+	ret = (ssize_t) write_double(buf, sizeof(buf), val);
+	if (!ret)
+		ret = iio_device_buffer_attr_write(dev, attr, buf);
+	return ret < 0 ? ret : 0;
+}
+
+int iio_device_buffer_attr_write_bool(const struct iio_device *dev,
+		const char *attr, bool val)
+{
+	ssize_t ret;
+
+	if (val)
+		ret = iio_device_buffer_attr_write(dev, attr, "1");
+	else
+		ret = iio_device_buffer_attr_write(dev, attr, "0");
+
+	return ret < 0 ? ret : 0;
+}
+
 ssize_t iio_device_debug_attr_read(const struct iio_device *dev,
 		const char *attr, char *dst, size_t len)
 {
 	if (dev->ctx->ops->read_device_attr)
 		return dev->ctx->ops->read_device_attr(dev,
-				attr, dst, len, true);
+				attr, dst, len, IIO_ATTR_TYPE_DEBUG);
 	else
 		return -ENOSYS;
 }
@@ -544,7 +740,7 @@ ssize_t iio_device_debug_attr_write_raw(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->write_device_attr)
 		return dev->ctx->ops->write_device_attr(dev,
-				attr, src, len, true);
+				attr, src, len, IIO_ATTR_TYPE_DEBUG);
 	else
 		return -ENOSYS;
 }
@@ -614,7 +810,7 @@ int iio_device_debug_attr_write_longlong(const struct iio_device *dev,
 	ssize_t ret;
 	char buf[1024];
 
-	snprintf(buf, sizeof(buf), "%lld", val);
+	iio_snprintf(buf, sizeof(buf), "%lld", val);
 	ret = iio_device_debug_attr_write(dev, attr, buf);
 
 	return ret < 0 ? ret : 0;
@@ -688,9 +884,10 @@ int iio_device_reg_write(struct iio_device *dev,
 		uint32_t address, uint32_t value)
 {
 	ssize_t ret;
-
 	char buf[1024];
-	snprintf(buf, sizeof(buf), "0x%" PRIx32 " 0x%" PRIx32, address, value);
+
+	iio_snprintf(buf, sizeof(buf), "0x%" PRIx32 " 0x%" PRIx32,
+			address, value);
 	ret = iio_device_debug_attr_write(dev, "direct_reg_access", buf);
 
 	return ret < 0 ? ret : 0;
@@ -716,12 +913,12 @@ int iio_device_reg_read(struct iio_device *dev,
 	return ret;
 }
 
-static int read_each_attr(struct iio_device *dev, bool is_debug,
+static int read_each_attr(struct iio_device *dev, enum iio_attr_type type,
 		int (*cb)(struct iio_device *dev,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	int ret;
+	int ret, buf_size;
 	char *buf, *ptr;
 	unsigned int i, count;
 
@@ -730,30 +927,67 @@ static int read_each_attr(struct iio_device *dev, bool is_debug,
 	if (!buf)
 		return -ENOMEM;
 
-	if (is_debug) {
-		count = iio_device_get_debug_attrs_count(dev);
-		ret = (int) iio_device_debug_attr_read(dev,
-				NULL, buf, 0x100000);
-	} else {
-		count = iio_device_get_attrs_count(dev);
-		ret = (int) iio_device_attr_read(dev, NULL, buf, 0x100000);
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			count = iio_device_get_attrs_count(dev);
+			ret = (int) iio_device_attr_read(dev,
+					NULL, buf, 0x100000);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			count = iio_device_get_debug_attrs_count(dev);
+			ret = (int) iio_device_debug_attr_read(dev,
+					NULL, buf, 0x100000);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			count = iio_device_get_buffer_attrs_count(dev);
+			ret = (int) iio_device_buffer_attr_read(dev,
+					NULL, buf, 0x100000);
+			break;
+		default:
+			ret = -EINVAL;
+			count = 0;
+			break;
 	}
 
 	if (ret < 0)
 		goto err_free_buf;
 
 	ptr = buf;
+	buf_size = ret;
 
 	for (i = 0; i < count; i++) {
 		const char *attr;
-		int32_t len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
+		int32_t len;
 
-		if (is_debug)
-			attr = iio_device_get_debug_attr(dev, i);
-		else
-			attr = iio_device_get_attr(dev, i);
+		if (buf_size < 4) {
+			ret = -EPROTO;
+			break;
+		}
 
+		len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
 		ptr += 4;
+		buf_size -= 4;
+
+		if (len > 0 && buf_size < len) {
+			ret = -EPROTO;
+			break;
+		}
+
+		switch(type){
+			case IIO_ATTR_TYPE_DEVICE:
+				attr = iio_device_get_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_DEBUG:
+				attr = iio_device_get_debug_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_BUFFER:
+				attr = iio_device_get_buffer_attr(dev, i);
+				break;
+			default:
+				attr = NULL;
+				break;
+		}
+
 		if (len > 0) {
 			ret = cb(dev, attr, ptr, (size_t) len, data);
 			if (ret < 0)
@@ -762,6 +996,10 @@ static int read_each_attr(struct iio_device *dev, bool is_debug,
 			if (len & 0x3)
 				len = ((len >> 2) + 1) << 2;
 			ptr += len;
+			if (len >= buf_size)
+				buf_size = 0;
+			else
+				buf_size -= len;
 		}
 	}
 
@@ -770,7 +1008,7 @@ err_free_buf:
 	return ret < 0 ? ret : 0;
 }
 
-static int write_each_attr(struct iio_device *dev, bool is_debug,
+static int write_each_attr(struct iio_device *dev, enum iio_attr_type type,
 		ssize_t (*cb)(struct iio_device *dev,
 			const char *attr, void *buf, size_t len, void *d),
 		void *data)
@@ -787,18 +1025,38 @@ static int write_each_attr(struct iio_device *dev, bool is_debug,
 
 	ptr = buf;
 
-	if (is_debug)
-		count = iio_device_get_debug_attrs_count(dev);
-	else
-		count = iio_device_get_attrs_count(dev);
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			count = iio_device_get_attrs_count(dev);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			count = iio_device_get_debug_attrs_count(dev);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			count = iio_device_get_buffer_attrs_count(dev);
+			break;
+		default:
+			ret = -EINVAL;
+			goto err_free_buf;
+	}
 
 	for (i = 0; i < count; i++) {
 		const char *attr;
 
-		if (is_debug)
-			attr = iio_device_get_debug_attr(dev, i);
-		else
-			attr = iio_device_get_attr(dev, i);
+		switch(type){
+			case IIO_ATTR_TYPE_DEVICE:
+				attr = iio_device_get_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_DEBUG:
+				attr = iio_device_get_debug_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_BUFFER:
+				attr = iio_device_get_buffer_attr(dev, i);
+				break;
+			default:
+				attr = NULL;
+				break;
+		}
 
 		ret = (int) cb(dev, attr, ptr + 4, len - 4, data);
 		if (ret < 0)
@@ -816,12 +1074,23 @@ static int write_each_attr(struct iio_device *dev, bool is_debug,
 		}
 	}
 
-	if (is_debug)
-		ret = (int) iio_device_debug_attr_write_raw(dev,
-				NULL, buf, ptr - buf);
-	else
-		ret = (int) iio_device_attr_write_raw(dev,
-				NULL, buf, ptr - buf);
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			ret = (int) iio_device_attr_write_raw(dev,
+					NULL, buf, ptr - buf);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			ret = (int) iio_device_debug_attr_write_raw(dev,
+					NULL, buf, ptr - buf);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			ret = (int) iio_device_buffer_attr_write_raw(dev,
+					NULL, buf, ptr - buf);
+			break;
+		default:
+			ret = -EINVAL;
+			break;
+	}
 
 err_free_buf:
 	free(buf);
@@ -833,7 +1102,15 @@ int iio_device_debug_attr_read_all(struct iio_device *dev,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	return read_each_attr(dev, true, cb, data);
+	return read_each_attr(dev, IIO_ATTR_TYPE_DEBUG, cb, data);
+}
+
+int iio_device_buffer_attr_read_all(struct iio_device *dev,
+		int (*cb)(struct iio_device *dev,
+			const char *attr, const char *val, size_t len, void *d),
+		void *data)
+{
+	return read_each_attr(dev, IIO_ATTR_TYPE_BUFFER, cb, data);
 }
 
 int iio_device_attr_read_all(struct iio_device *dev,
@@ -841,7 +1118,7 @@ int iio_device_attr_read_all(struct iio_device *dev,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	return read_each_attr(dev, false, cb, data);
+	return read_each_attr(dev, IIO_ATTR_TYPE_DEVICE ,cb, data);
 }
 
 int iio_device_debug_attr_write_all(struct iio_device *dev,
@@ -849,7 +1126,15 @@ int iio_device_debug_attr_write_all(struct iio_device *dev,
 			const char *attr, void *buf, size_t len, void *d),
 		void *data)
 {
-	return write_each_attr(dev, true, cb, data);
+	return write_each_attr(dev, IIO_ATTR_TYPE_DEBUG, cb, data);
+}
+
+int iio_device_buffer_attr_write_all(struct iio_device *dev,
+		ssize_t (*cb)(struct iio_device *dev,
+			const char *attr, void *buf, size_t len, void *d),
+		void *data)
+{
+	return write_each_attr(dev, IIO_ATTR_TYPE_BUFFER, cb, data);
 }
 
 int iio_device_attr_write_all(struct iio_device *dev,
@@ -857,7 +1142,7 @@ int iio_device_attr_write_all(struct iio_device *dev,
 			const char *attr, void *buf, size_t len, void *d),
 		void *data)
 {
-	return write_each_attr(dev, false, cb, data);
+	return write_each_attr(dev, IIO_ATTR_TYPE_DEVICE, cb, data);
 }
 
 const struct iio_context * iio_device_get_context(const struct iio_device *dev)
