@@ -18,6 +18,7 @@
 
 #include "debug.h"
 #include "iio-private.h"
+#include "sort.h"
 
 #include <dirent.h>
 #include <errno.h>
@@ -38,6 +39,9 @@
 #include <time.h>
 #include <unistd.h>
 #include <fcntl.h>
+#ifdef WITH_LOCAL_CONFIG
+#include <ini.h>
+#endif
 
 #define DEFAULT_TIMEOUT_MS 1000
 
@@ -53,11 +57,11 @@
 
 /* Forward declarations */
 static ssize_t local_read_dev_attr(const struct iio_device *dev,
-		const char *attr, char *dst, size_t len, bool is_debug);
+		const char *attr, char *dst, size_t len, enum iio_attr_type type);
 static ssize_t local_read_chn_attr(const struct iio_channel *chn,
 		const char *attr, char *dst, size_t len);
 static ssize_t local_write_dev_attr(const struct iio_device *dev,
-		const char *attr, const char *src, size_t len, bool is_debug);
+		const char *attr, const char *src, size_t len, enum iio_attr_type type);
 static ssize_t local_write_chn_attr(const struct iio_channel *chn,
 		const char *attr, const char *src, size_t len);
 
@@ -85,7 +89,9 @@ struct iio_context_pdata {
 struct iio_device_pdata {
 	int fd;
 	bool blocking;
-	unsigned int samples_count, nb_blocks;
+	unsigned int samples_count;
+	unsigned int max_nb_blocks;
+	unsigned int allocated_nb_blocks;
 
 	struct block *blocks;
 	void **addrs;
@@ -95,11 +101,22 @@ struct iio_device_pdata {
 	int cancel_fd;
 };
 
+struct iio_channel_pdata {
+	char *enable_fn;
+	struct iio_channel_attr *protected_attrs;
+	unsigned int nb_protected_attrs;
+};
+
 static const char * const device_attrs_blacklist[] = {
 	"dev",
 	"uevent",
 };
 
+static const char * const buffer_attrs_reserved[] = {
+	"length",
+	"enable",
+};
+
 static int ioctl_nointr(int fd, unsigned long request, void *data)
 {
 	int ret;
@@ -111,9 +128,22 @@ static int ioctl_nointr(int fd, unsigned long request, void *data)
 	return ret;
 }
 
+static void local_free_channel_pdata(struct iio_channel *chn)
+{
+	if (chn->pdata) {
+		free(chn->pdata->enable_fn);
+		free(chn->pdata);
+	}
+}
+
 static void local_free_pdata(struct iio_device *device)
 {
-	if (device && device->pdata) {
+	unsigned int i;
+
+	for (i = 0; i < device->nb_channels; i++)
+		local_free_channel_pdata(device->channels[i]);
+
+	if (device->pdata) {
 		free(device->pdata->blocks);
 		free(device->pdata->addrs);
 		free(device->pdata);
@@ -125,8 +155,12 @@ static void local_shutdown(struct iio_context *ctx)
 	/* Free the backend data stored in every device structure */
 	unsigned int i;
 
-	for (i = 0; i < ctx->nb_devices; i++)
-		local_free_pdata(ctx->devices[i]);
+	for (i = 0; i < ctx->nb_devices; i++) {
+		struct iio_device *dev = ctx->devices[i];
+
+		iio_device_close(dev);
+		local_free_pdata(dev);
+	}
 
 	free(ctx->pdata);
 }
@@ -143,15 +177,19 @@ static void strcut(char *str, int nb)
 
 static int set_channel_name(struct iio_channel *chn)
 {
+	struct iio_channel_pdata *pdata = chn->pdata;
 	size_t prefix_len = 0;
 	const char *attr0;
 	const char *ptr;
 	unsigned int i;
 
-	if (chn->nb_attrs < 2)
+	if (chn->nb_attrs + pdata->nb_protected_attrs < 2)
 		return 0;
 
-	attr0 = ptr = chn->attrs[0].name;
+	if (chn->nb_attrs)
+		attr0 = ptr = chn->attrs[0].name;
+	else
+		attr0 = ptr = pdata->protected_attrs[0].name;
 
 	while (true) {
 		bool can_fix = true;
@@ -165,6 +203,12 @@ static int set_channel_name(struct iio_channel *chn)
 		for (i = 1; can_fix && i < chn->nb_attrs; i++)
 			can_fix = !strncmp(attr0, chn->attrs[i].name, len);
 
+		for (i = !chn->nb_attrs;
+				can_fix && i < pdata->nb_protected_attrs; i++) {
+			can_fix = !strncmp(attr0,
+					pdata->protected_attrs[i].name, len);
+		}
+
 		if (!can_fix)
 			break;
 
@@ -186,6 +230,8 @@ static int set_channel_name(struct iio_channel *chn)
 		/* Shrink the attribute name */
 		for (i = 0; i < chn->nb_attrs; i++)
 			strcut(chn->attrs[i].name, prefix_len);
+		for (i = 0; i < pdata->nb_protected_attrs; i++)
+			strcut(pdata->protected_attrs[i].name, prefix_len);
 	}
 
 	return 0;
@@ -301,7 +347,7 @@ static ssize_t local_read(const struct iio_device *dev,
 		if (ret == -1) {
 			if (pdata->blocking && errno == EAGAIN)
 				continue;
-			ret = -EIO;
+			ret = -errno;
 			break;
 		} else if (ret == 0) {
 			ret = -EIO;
@@ -349,7 +395,7 @@ static ssize_t local_write(const struct iio_device *dev,
 			if (pdata->blocking && errno == EAGAIN)
 				continue;
 
-			ret = -EIO;
+			ret = -errno;
 			break;
 		} else if (ret == 0) {
 			ret = -EIO;
@@ -379,7 +425,7 @@ static ssize_t local_enable_buffer(const struct iio_device *dev)
 			pdata->buffer_enabled = true;
 	}
 
-	return 0;
+	return ret;
 }
 
 static int local_set_kernel_buffers_count(const struct iio_device *dev,
@@ -390,7 +436,7 @@ static int local_set_kernel_buffers_count(const struct iio_device *dev,
 	if (pdata->fd != -1)
 		return -EBUSY;
 
-	pdata->nb_blocks = nb_blocks;
+	pdata->max_nb_blocks = nb_blocks;
 
 	return 0;
 }
@@ -437,6 +483,8 @@ static ssize_t local_get_buffer(const struct iio_device *dev,
 			*addr_ptr = pdata->addrs[pdata->last_dequeued];
 			return (ssize_t) last_block->bytes_used;
 		}
+
+		pdata->last_dequeued = -1;
 	}
 
 	clock_gettime(CLOCK_MONOTONIC, &start);
@@ -470,16 +518,34 @@ static ssize_t local_get_buffer(const struct iio_device *dev,
 }
 
 static ssize_t local_read_all_dev_attrs(const struct iio_device *dev,
-		char *dst, size_t len, bool is_debug)
+		char *dst, size_t len, enum iio_attr_type type)
 {
-	unsigned int i, nb = is_debug ? dev->nb_debug_attrs : dev->nb_attrs;
-	char **attrs = is_debug ? dev->debug_attrs : dev->attrs;
+	unsigned int i, nb;
+	char **attrs;
 	char *ptr = dst;
 
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			nb =  dev->nb_attrs;
+			attrs = dev->attrs;
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			nb =  dev->nb_debug_attrs;
+			attrs = dev->debug_attrs;
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			nb =  dev->nb_buffer_attrs;
+			attrs = dev->buffer_attrs;
+			break;
+		default:
+			return -EINVAL;
+			break;
+	}
+
 	for (i = 0; len >= 4 && i < nb; i++) {
 		/* Recursive! */
 		ssize_t ret = local_read_dev_attr(dev, attrs[i],
-				ptr + 4, len - 4, is_debug);
+				ptr + 4, len - 4, type);
 		*(uint32_t *) ptr = iio_htobe32(ret);
 
 		/* Align the length to 4 bytes */
@@ -543,12 +609,30 @@ static int local_buffer_analyze(unsigned int nb, const char *src, size_t len)
 }
 
 static ssize_t local_write_all_dev_attrs(const struct iio_device *dev,
-		const char *src, size_t len, bool is_debug)
+		const char *src, size_t len, enum iio_attr_type type)
 {
-	unsigned int i, nb = is_debug ? dev->nb_debug_attrs : dev->nb_attrs;
-	char **attrs = is_debug ? dev->debug_attrs : dev->attrs;
+	unsigned int i, nb;
+	char **attrs;
 	const char *ptr = src;
 
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			nb =  dev->nb_attrs;
+			attrs = dev->attrs;
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			nb =  dev->nb_debug_attrs;
+			attrs = dev->debug_attrs;
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			nb =  dev->nb_buffer_attrs;
+			attrs = dev->buffer_attrs;
+			break;
+		default:
+			return -EINVAL;
+			break;
+	}
+
 	/* First step: Verify that the buffer is in the correct format */
 	if (local_buffer_analyze(nb, src, len))
 		return -EINVAL;
@@ -559,7 +643,7 @@ static ssize_t local_write_all_dev_attrs(const struct iio_device *dev,
 		ptr += 4;
 
 		if (val > 0) {
-			local_write_dev_attr(dev, attrs[i], ptr, val, is_debug);
+			local_write_dev_attr(dev, attrs[i], ptr, val, type);
 
 			/* Align the length to 4 bytes */
 			if (val & 3)
@@ -600,21 +684,32 @@ static ssize_t local_write_all_chn_attrs(const struct iio_channel *chn,
 }
 
 static ssize_t local_read_dev_attr(const struct iio_device *dev,
-		const char *attr, char *dst, size_t len, bool is_debug)
+		const char *attr, char *dst, size_t len, enum iio_attr_type type)
 {
 	FILE *f;
 	char buf[1024];
 	ssize_t ret;
 
 	if (!attr)
-		return local_read_all_dev_attrs(dev, dst, len, is_debug);
+		return local_read_all_dev_attrs(dev, dst, len, type);
+
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			iio_snprintf(buf, sizeof(buf), "/sys/bus/iio/devices/%s/%s",
+					dev->id, attr);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			iio_snprintf(buf, sizeof(buf), "/sys/kernel/debug/iio/%s/%s",
+					dev->id, attr);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			iio_snprintf(buf, sizeof(buf), "/sys/bus/iio/devices/%s/buffer/%s",
+					dev->id, attr);
+			break;
+		default:
+			return -EINVAL;
+	}
 
-	if (is_debug)
-		snprintf(buf, sizeof(buf), "/sys/kernel/debug/iio/%s/%s",
-				dev->id, attr);
-	else
-		snprintf(buf, sizeof(buf), "/sys/bus/iio/devices/%s/%s",
-				dev->id, attr);
 	f = fopen(buf, "re");
 	if (!f)
 		return -errno;
@@ -630,21 +725,32 @@ static ssize_t local_read_dev_attr(const struct iio_device *dev,
 }
 
 static ssize_t local_write_dev_attr(const struct iio_device *dev,
-		const char *attr, const char *src, size_t len, bool is_debug)
+		const char *attr, const char *src, size_t len, enum iio_attr_type type)
 {
 	FILE *f;
 	char buf[1024];
 	ssize_t ret;
 
 	if (!attr)
-		return local_write_all_dev_attrs(dev, src, len, is_debug);
+		return local_write_all_dev_attrs(dev, src, len, type);
+
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			iio_snprintf(buf, sizeof(buf), "/sys/bus/iio/devices/%s/%s",
+					dev->id, attr);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			iio_snprintf(buf, sizeof(buf), "/sys/kernel/debug/iio/%s/%s",
+					dev->id, attr);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			iio_snprintf(buf, sizeof(buf), "/sys/bus/iio/devices/%s/buffer/%s",
+					dev->id, attr);
+			break;
+		default:
+			return -EINVAL;
+	}
 
-	if (is_debug)
-		snprintf(buf, sizeof(buf), "/sys/kernel/debug/iio/%s/%s",
-				dev->id, attr);
-	else
-		snprintf(buf, sizeof(buf), "/sys/bus/iio/devices/%s/%s",
-				dev->id, attr);
 	f = fopen(buf, "we");
 	if (!f)
 		return -errno;
@@ -687,10 +793,16 @@ static ssize_t local_write_chn_attr(const struct iio_channel *chn,
 	return local_write_dev_attr(chn->dev, attr, src, len, false);
 }
 
-static int channel_write_state(const struct iio_channel *chn)
+static int channel_write_state(const struct iio_channel *chn, bool en)
 {
-	char *en = iio_channel_is_enabled(chn) ? "1" : "0";
-	ssize_t ret = local_write_chn_attr(chn, "en", en, 2);
+	ssize_t ret;
+
+	if (!chn->pdata->enable_fn) {
+		ERROR("Libiio bug: No \"en\" attribute parsed\n");
+		return -EINVAL;
+	}
+
+	ret = local_write_chn_attr(chn, chn->pdata->enable_fn, en ? "1" : "0", 2);
 	if (ret < 0)
 		return (int) ret;
 	else
@@ -701,23 +813,34 @@ static int enable_high_speed(const struct iio_device *dev)
 {
 	struct block_alloc_req req;
 	struct iio_device_pdata *pdata = dev->pdata;
+	unsigned int nb_blocks;
 	unsigned int i;
 	int ret, fd = pdata->fd;
 
+	/*
+	 * For the BLOCK_ALLOC_IOCTL ioctl it is not possible to distingush
+	 * between an error during the allocation (e.g. incorrect size) or
+	 * whether the high-speed interface is not supported. BLOCK_FREE_IOCTL does
+	 * never fail if the device supports the high-speed interface, so we use it
+	 * here. Calling it when no blocks are allocated the ioctl has no effect.
+	 */
+	ret = ioctl_nointr(fd, BLOCK_FREE_IOCTL, NULL);
+	if (ret < 0)
+		return -ENOSYS;
+
 	if (pdata->cyclic) {
-		pdata->nb_blocks = 1;
+		nb_blocks = 1;
 		DEBUG("Enabling cyclic mode\n");
 	} else {
+		nb_blocks = pdata->max_nb_blocks;
 		DEBUG("Cyclic mode not enabled\n");
 	}
 
-	pdata->blocks = calloc(pdata->nb_blocks, sizeof(*pdata->blocks));
-	if (!pdata->blocks) {
-		pdata->nb_blocks = 0;
+	pdata->blocks = calloc(nb_blocks, sizeof(*pdata->blocks));
+	if (!pdata->blocks)
 		return -ENOMEM;
-	}
 
-	pdata->addrs = calloc(pdata->nb_blocks, sizeof(*pdata->addrs));
+	pdata->addrs = calloc(nb_blocks, sizeof(*pdata->addrs));
 	if (!pdata->addrs) {
 		free(pdata->blocks);
 		pdata->blocks = NULL;
@@ -728,7 +851,7 @@ static int enable_high_speed(const struct iio_device *dev)
 	req.type = 0;
 	req.size = pdata->samples_count *
 		iio_device_get_sample_size_mask(dev, dev->mask, dev->words);
-	req.count = pdata->nb_blocks;
+	req.count = nb_blocks;
 
 	ret = ioctl_nointr(fd, BLOCK_ALLOC_IOCTL, &req);
 	if (ret < 0) {
@@ -736,11 +859,16 @@ static int enable_high_speed(const struct iio_device *dev)
 		goto err_freemem;
 	}
 
+	if (req.count == 0) {
+		ret = -ENOMEM;
+		goto err_block_free;
+	}
+
 	/* We might get less blocks than what we asked for */
-	pdata->nb_blocks = req.count;
+	pdata->allocated_nb_blocks = req.count;
 
 	/* mmap all the blocks */
-	for (i = 0; i < pdata->nb_blocks; i++) {
+	for (i = 0; i < pdata->allocated_nb_blocks; i++) {
 		pdata->blocks[i].id = i;
 		ret = ioctl_nointr(fd, BLOCK_QUERY_IOCTL, &pdata->blocks[i]);
 		if (ret) {
@@ -769,7 +897,9 @@ static int enable_high_speed(const struct iio_device *dev)
 err_munmap:
 	for (; i > 0; i--)
 		munmap(pdata->addrs[i - 1], pdata->blocks[i - 1].size);
+err_block_free:
 	ioctl_nointr(fd, BLOCK_FREE_IOCTL, 0);
+	pdata->allocated_nb_blocks = 0;
 err_freemem:
 	free(pdata->addrs);
 	pdata->addrs = NULL;
@@ -793,7 +923,7 @@ static int local_open(const struct iio_device *dev,
 	if (ret < 0)
 		return ret;
 
-	snprintf(buf, sizeof(buf), "%lu", (unsigned long) samples_count);
+	iio_snprintf(buf, sizeof(buf), "%lu", (unsigned long) samples_count);
 	ret = local_write_dev_attr(dev, "buffer/length",
 			buf, strlen(buf) + 1, false);
 	if (ret < 0)
@@ -803,7 +933,7 @@ static int local_open(const struct iio_device *dev,
 	if (pdata->cancel_fd == -1)
 		return -errno;
 
-	snprintf(buf, sizeof(buf), "/dev/%s", dev->id);
+	iio_snprintf(buf, sizeof(buf), "/dev/%s", dev->id);
 	pdata->fd = open(buf, O_RDWR | O_CLOEXEC | O_NONBLOCK);
 	if (pdata->fd == -1) {
 		ret = -errno;
@@ -814,7 +944,7 @@ static int local_open(const struct iio_device *dev,
 	for (i = 0; i < dev->nb_channels; i++) {
 		struct iio_channel *chn = dev->channels[i];
 		if (chn->index >= 0 && !iio_channel_is_enabled(chn)) {
-			ret = channel_write_state(chn);
+			ret = channel_write_state(chn, false);
 			if (ret < 0)
 				goto err_close;
 		}
@@ -823,7 +953,7 @@ static int local_open(const struct iio_device *dev,
 	for (i = 0; i < dev->nb_channels; i++) {
 		struct iio_channel *chn = dev->channels[i];
 		if (chn->index >= 0 && iio_channel_is_enabled(chn)) {
-			ret = channel_write_state(chn);
+			ret = channel_write_state(chn, true);
 			if (ret < 0)
 				goto err_close;
 		}
@@ -833,20 +963,27 @@ static int local_open(const struct iio_device *dev,
 	pdata->cyclic_buffer_enqueued = false;
 	pdata->buffer_enabled = false;
 	pdata->samples_count = samples_count;
-	pdata->is_high_speed = !enable_high_speed(dev);
+
+	ret = enable_high_speed(dev);
+	if (ret < 0 && ret != -ENOSYS)
+		goto err_close;
+
+	pdata->is_high_speed = !ret;
 
 	if (!pdata->is_high_speed) {
-		unsigned long size = samples_count * pdata->nb_blocks;
+		unsigned long size = samples_count * pdata->max_nb_blocks;
 		WARNING("High-speed mode not enabled\n");
 
 		/* Cyclic mode is only supported in high-speed mode */
-		if (cyclic)
-			return -EPERM;
+		if (cyclic) {
+			ret = -EPERM;
+			goto err_close;
+		}
 
 		/* Increase the size of the kernel buffer, when using the
 		 * low-speed interface. This avoids losing samples when
 		 * refilling the iio_buffer. */
-		snprintf(buf, sizeof(buf), "%lu", size);
+		iio_snprintf(buf, sizeof(buf), "%lu", size);
 		ret = local_write_dev_attr(dev, "buffer/length",
 				buf, strlen(buf) + 1, false);
 		if (ret < 0)
@@ -870,6 +1007,7 @@ err_close_cancel_fd:
 static int local_close(const struct iio_device *dev)
 {
 	struct iio_device_pdata *pdata = dev->pdata;
+	unsigned int i;
 	int ret;
 
 	if (pdata->fd == -1)
@@ -877,9 +1015,10 @@ static int local_close(const struct iio_device *dev)
 
 	if (pdata->is_high_speed) {
 		unsigned int i;
-		for (i = 0; i < pdata->nb_blocks; i++)
+		for (i = 0; i < pdata->allocated_nb_blocks; i++)
 			munmap(pdata->addrs[i], pdata->blocks[i].size);
 		ioctl_nointr(pdata->fd, BLOCK_FREE_IOCTL, 0);
+		pdata->allocated_nb_blocks = 0;
 		free(pdata->addrs);
 		pdata->addrs = NULL;
 		free(pdata->blocks);
@@ -896,6 +1035,14 @@ static int local_close(const struct iio_device *dev)
 	pdata->cancel_fd = -1;
 
 	ret = local_write_dev_attr(dev, "buffer/enable", "0", 2, false);
+
+	for (i = 0; i < dev->nb_channels; i++) {
+		struct iio_channel *chn = dev->channels[i];
+
+		if (chn->pdata->enable_fn)
+			channel_write_state(chn, false);
+	}
+
 	return (ret < 0) ? ret : 0;
 }
 
@@ -1012,11 +1159,11 @@ static char * get_short_attr_name(struct iio_channel *chn, const char *attr)
 
 	if (chn->name) {
 		size_t len = strlen(chn->name);
-		if  (strncmp(chn->name, ptr, len) == 0 && ptr[len] == '_')
+		if (strncmp(chn->name, ptr, len) == 0 && ptr[len] == '_')
 			ptr += len + 1;
 	}
 
-	return strdup(ptr);
+	return iio_strdup(ptr);
 }
 
 static int read_device_name(struct iio_device *dev)
@@ -1028,7 +1175,7 @@ static int read_device_name(struct iio_device *dev)
 	else if (ret == 0)
 		return -EIO;
 
-	dev->name = strdup(buf);
+	dev->name = iio_strdup(buf);
 	if (!dev->name)
 		return -ENOMEM;
 	else
@@ -1047,7 +1194,7 @@ static int add_attr_to_device(struct iio_device *dev, const char *attr)
 	if (!strcmp(attr, "name"))
 		return read_device_name(dev);
 
-	name = strdup(attr);
+	name = iio_strdup(attr);
 	if (!name)
 		return -ENOMEM;
 
@@ -1063,18 +1210,128 @@ static int add_attr_to_device(struct iio_device *dev, const char *attr)
 	return 0;
 }
 
+static int handle_protected_scan_element_attr(struct iio_channel *chn,
+			const char *name, const char *path)
+{
+	struct iio_device *dev = chn->dev;
+	char buf[1024];
+	int ret;
+
+	if (!strcmp(name, "index")) {
+		ret = local_read_dev_attr(dev, path, buf, sizeof(buf), false);
+		if (ret > 0)
+			chn->index = atol(buf);
+
+	} else if (!strcmp(name, "type")) {
+		ret = local_read_dev_attr(dev, path, buf, sizeof(buf), false);
+		if (ret > 0) {
+			char endian, sign;
+
+			if (strchr(buf, 'X')) {
+				sscanf(buf, "%ce:%c%u/%uX%u>>%u", &endian, &sign,
+					&chn->format.bits, &chn->format.length,
+					&chn->format.repeat, &chn->format.shift);
+			} else {
+				chn->format.repeat = 1;
+				sscanf(buf, "%ce:%c%u/%u>>%u", &endian, &sign,
+					&chn->format.bits, &chn->format.length,
+					&chn->format.shift);
+			}
+			chn->format.is_signed = (sign == 's' || sign == 'S');
+			chn->format.is_fully_defined =
+					(sign == 'S' || sign == 'U'||
+					chn->format.bits == chn->format.length);
+			chn->format.is_be = endian == 'b';
+		}
+
+	} else if (!strcmp(name, "en")) {
+		if (chn->pdata->enable_fn) {
+			ERROR("Libiio bug: \"en\" attribute already parsed for channel %s!\n",
+					chn->id);
+			return -EINVAL;
+		}
+
+		chn->pdata->enable_fn = iio_strdup(path);
+		if (!chn->pdata->enable_fn)
+			return -ENOMEM;
+
+	} else {
+		return -EINVAL;
+	}
+
+	return 0;
+}
+
+static int handle_scan_elements(struct iio_channel *chn)
+{
+	struct iio_channel_pdata *pdata = chn->pdata;
+	unsigned int i;
+
+	for (i = 0; i < pdata->nb_protected_attrs; i++) {
+		int ret = handle_protected_scan_element_attr(chn,
+				pdata->protected_attrs[i].name,
+				pdata->protected_attrs[i].filename);
+		if (ret < 0)
+			return ret;
+	}
+
+	return 0;
+}
+
+static int add_protected_attr(struct iio_channel *chn, char *name, char *fn)
+{
+	struct iio_channel_pdata *pdata = chn->pdata;
+	struct iio_channel_attr *attrs;
+
+	attrs = realloc(pdata->protected_attrs,
+			(1 + pdata->nb_protected_attrs) * sizeof(*attrs));
+	if (!attrs)
+		return -ENOMEM;
+
+	attrs[pdata->nb_protected_attrs].name = name;
+	attrs[pdata->nb_protected_attrs++].filename = fn;
+	pdata->protected_attrs = attrs;
+
+	DEBUG("Add protected attr \'%s\' to channel \'%s\'\n", name, chn->id);
+	return 0;
+}
+
+static void free_protected_attrs(struct iio_channel *chn)
+{
+	struct iio_channel_pdata *pdata = chn->pdata;
+	unsigned int i;
+
+	for (i = 0; i < pdata->nb_protected_attrs; i++) {
+		free(pdata->protected_attrs[i].name);
+		free(pdata->protected_attrs[i].filename);
+	}
+
+	free(pdata->protected_attrs);
+	pdata->nb_protected_attrs = 0;
+	pdata->protected_attrs = NULL;
+}
+
 static int add_attr_to_channel(struct iio_channel *chn,
-		const char *attr, const char *path)
+		const char *attr, const char *path, bool is_scan_element)
 {
 	struct iio_channel_attr *attrs;
 	char *fn, *name = get_short_attr_name(chn, attr);
 	if (!name)
 		return -ENOMEM;
 
-	fn = strdup(path);
+	fn = iio_strdup(path);
 	if (!fn)
 		goto err_free_name;
 
+	if (is_scan_element) {
+		int ret = add_protected_attr(chn, name, fn);
+
+		if (ret < 0)
+			goto err_free_fn;
+
+		return 0;
+	}
+
 	attrs = realloc(chn->attrs, (1 + chn->nb_attrs) *
 			sizeof(struct iio_channel_attr));
 	if (!attrs)
@@ -1103,7 +1360,9 @@ static int add_channel_to_device(struct iio_device *dev,
 
 	channels[dev->nb_channels++] = chn;
 	dev->channels = channels;
-	DEBUG("Added channel \'%s\' to device \'%s\'\n", chn->id, dev->id);
+	DEBUG("Added %s channel \'%s\' to device \'%s\'\n",
+		chn->is_output ? "output" : "input", chn->id, dev->id);
+
 	return 0;
 }
 
@@ -1122,23 +1381,33 @@ static int add_device_to_context(struct iio_context *ctx,
 }
 
 static struct iio_channel *create_channel(struct iio_device *dev,
-		char *id, const char *attr, const char *path)
+		char *id, const char *attr, const char *path,
+		bool is_scan_element)
 {
 	struct iio_channel *chn = zalloc(sizeof(*chn));
 	if (!chn)
 		return NULL;
 
+	chn->pdata = zalloc(sizeof(*chn->pdata));
+	if (!chn->pdata)
+		goto err_free_chn;
+
 	if (!strncmp(attr, "out_", 4))
 		chn->is_output = true;
 	else if (strncmp(attr, "in_", 3))
-		goto err_free_chn;
+		goto err_free_chn_pdata;
 
 	chn->dev = dev;
 	chn->id = id;
+	chn->is_scan_element = is_scan_element;
+	chn->index = -ENOENT;
 
-	if (!add_attr_to_channel(chn, attr, path))
+	if (!add_attr_to_channel(chn, attr, path, is_scan_element))
 		return chn;
 
+err_free_chn_pdata:
+	free(chn->pdata->enable_fn);
+	free(chn->pdata);
 err_free_chn:
 	free(chn);
 	return NULL;
@@ -1161,13 +1430,14 @@ static int add_channel(struct iio_device *dev, const char *name,
 		if (!strcmp(chn->id, channel_id)
 				&& chn->is_output == (name[0] == 'o')) {
 			free(channel_id);
-			ret = add_attr_to_channel(chn, name, path);
+			ret = add_attr_to_channel(chn, name, path,
+					dir_is_scan_elements);
 			chn->is_scan_element = dir_is_scan_elements && !ret;
 			return ret;
 		}
 	}
 
-	chn = create_channel(dev, channel_id, name, path);
+	chn = create_channel(dev, channel_id, name, path, dir_is_scan_elements);
 	if (!chn) {
 		free(channel_id);
 		return -ENXIO;
@@ -1176,10 +1446,11 @@ static int add_channel(struct iio_device *dev, const char *name,
 	iio_channel_init_finalize(chn);
 
 	ret = add_channel_to_device(dev, chn);
-	if (ret)
+	if (ret) {
+		free(chn->pdata->enable_fn);
+		free(chn->pdata);
 		free_channel(chn);
-	else
-		chn->is_scan_element = dir_is_scan_elements;
+	}
 	return ret;
 }
 
@@ -1240,7 +1511,7 @@ static int detect_global_attr(struct iio_device *dev, const char *attr,
 		if (is_global_attr(chn, attr) == level) {
 			int ret;
 			*match = true;
-			ret = add_attr_to_channel(chn, attr, attr);
+			ret = add_attr_to_channel(chn, attr, attr, false);
 			if (ret)
 				return ret;
 		}
@@ -1307,6 +1578,33 @@ static int detect_and_move_global_attrs(struct iio_device *dev)
 	return 0;
 }
 
+static int add_buffer_attr(void *d, const char *path)
+{
+	struct iio_device *dev = (struct iio_device *) d;
+	const char *name = strrchr(path, '/') + 1;
+	char **attrs, *attr;
+	int i;
+
+	for (i = 0; i < ARRAY_SIZE(buffer_attrs_reserved); i++)
+		if (!strcmp(buffer_attrs_reserved[i], name))
+			return 0;
+
+	attr = iio_strdup(name);
+	if (!attr)
+		return -ENOMEM;
+
+	attrs = realloc(dev->buffer_attrs, (1 + dev->nb_buffer_attrs) * sizeof(char *));
+	if (!attrs) {
+		free(attr);
+		return -ENOMEM;
+	}
+
+	attrs[dev->nb_buffer_attrs++] = attr;
+	dev->buffer_attrs = attrs;
+	DEBUG("Added buffer attr \'%s\' to device \'%s\'\n", attr, dev->id);
+	return 0;
+}
+
 static int add_attr_or_channel_helper(struct iio_device *dev,
 		const char *path, bool dir_is_scan_elements)
 {
@@ -1314,7 +1612,7 @@ static int add_attr_or_channel_helper(struct iio_device *dev,
 	const char *name = strrchr(path, '/') + 1;
 
 	if (dir_is_scan_elements) {
-		snprintf(buf, sizeof(buf), "scan_elements/%s", name);
+		iio_snprintf(buf, sizeof(buf), "scan_elements/%s", name);
 		path = buf;
 	} else {
 		if (!is_channel(name, true))
@@ -1338,43 +1636,36 @@ static int add_scan_element(void *d, const char *path)
 static int foreach_in_dir(void *d, const char *path, bool is_dir,
 		int (*callback)(void *, const char *))
 {
-	long name_max;
-	struct dirent *entry, *result;
-	DIR *dir = opendir(path);
+	struct dirent *entry;
+	DIR *dir;
+	int ret = 0;
+
+	dir = opendir(path);
 	if (!dir)
 		return -errno;
 
-	name_max = pathconf(path, _PC_NAME_MAX);
-	if (name_max == -1)
-		name_max = 255;
-	entry = malloc(offsetof(struct dirent, d_name) + name_max + 1);
-	if (!entry) {
-		closedir(dir);
-		return -ENOMEM;
-	}
-
 	while (true) {
 		struct stat st;
 		char buf[1024];
-		int ret = readdir_r(dir, entry, &result);
-		if (ret) {
-			iio_strerror(ret, buf, sizeof(buf));
-			ERROR("Unable to open directory %s: %s\n", path, buf);
-			free(entry);
-			closedir(dir);
-			return -ret;
-		}
-		if (!result)
-			break;
 
-		snprintf(buf, sizeof(buf), "%s/%s", path, entry->d_name);
+		errno = 0;
+		entry = readdir(dir);
+		if (!entry) {
+			if (!errno)
+				break;
+
+			ret = -errno;
+			iio_strerror(errno, buf, sizeof(buf));
+			ERROR("Unable to open directory %s: %s\n", path, buf);
+			goto out_close_dir;
+		}
+
+		iio_snprintf(buf, sizeof(buf), "%s/%s", path, entry->d_name);
 		if (stat(buf, &st) < 0) {
 			ret = -errno;
 			iio_strerror(errno, buf, sizeof(buf));
 			ERROR("Unable to stat file: %s\n", buf);
-			free(entry);
-			closedir(dir);
-			return ret;
+			goto out_close_dir;
 		}
 
 		if (is_dir && S_ISDIR(st.st_mode) && entry->d_name[0] != '.')
@@ -1384,23 +1675,21 @@ static int foreach_in_dir(void *d, const char *path, bool is_dir,
 		else
 			continue;
 
-		if (ret < 0) {
-			free(entry);
-			closedir(dir);
-			return ret;
-		}
+		if (ret < 0)
+			goto out_close_dir;
 	}
 
-	free(entry);
+out_close_dir:
 	closedir(dir);
-	return 0;
+	return ret;
 }
 
 static int add_scan_elements(struct iio_device *dev, const char *devpath)
 {
 	struct stat st;
 	char buf[1024];
-	snprintf(buf, sizeof(buf), "%s/scan_elements", devpath);
+
+	iio_snprintf(buf, sizeof(buf), "%s/scan_elements", devpath);
 
 	if (!stat(buf, &st) && S_ISDIR(st.st_mode)) {
 		int ret = foreach_in_dir(dev, buf, false, add_scan_element);
@@ -1411,6 +1700,25 @@ static int add_scan_elements(struct iio_device *dev, const char *devpath)
 	return 0;
 }
 
+static int add_buffer_attributes(struct iio_device *dev, const char *devpath)
+{
+	struct stat st;
+	char buf[1024];
+
+	iio_snprintf(buf, sizeof(buf), "%s/buffer", devpath);
+
+	if (!stat(buf, &st) && S_ISDIR(st.st_mode)) {
+		int ret = foreach_in_dir(dev, buf, false, add_buffer_attr);
+		if (ret < 0)
+			return ret;
+
+		qsort(dev->buffer_attrs, dev->nb_buffer_attrs, sizeof(char *),
+			iio_buffer_attr_compare);
+	}
+
+	return 0;
+}
+
 static int create_device(void *d, const char *path)
 {
 	uint32_t *mask = NULL;
@@ -1429,10 +1737,10 @@ static int create_device(void *d, const char *path)
 
 	dev->pdata->fd = -1;
 	dev->pdata->blocking = true;
-	dev->pdata->nb_blocks = NB_BLOCKS;
+	dev->pdata->max_nb_blocks = NB_BLOCKS;
 
 	dev->ctx = ctx;
-	dev->id = strdup(strrchr(path, '/') + 1);
+	dev->id = iio_strdup(strrchr(path, '/') + 1);
 	if (!dev->id) {
 		local_free_pdata(dev);
 		free(dev);
@@ -1440,40 +1748,61 @@ static int create_device(void *d, const char *path)
 	}
 
 	ret = foreach_in_dir(dev, path, false, add_attr_or_channel);
-	if (ret < 0) {
-		free_device(dev);
-		return ret;
-	}
+	if (ret < 0)
+		goto err_free_device;
+
+	ret = add_buffer_attributes(dev, path);
+	if (ret < 0)
+		goto err_free_device;
 
 	ret = add_scan_elements(dev, path);
-	if (ret < 0) {
-		free_device(dev);
-		return ret;
-	}
+	if (ret < 0)
+		goto err_free_scan_elements;
 
-	for (i = 0; i < dev->nb_channels; i++)
-		set_channel_name(dev->channels[i]);
+	for (i = 0; i < dev->nb_channels; i++) {
+		struct iio_channel *chn = dev->channels[i];
+
+		set_channel_name(chn);
+		ret = handle_scan_elements(chn);
+		free_protected_attrs(chn);
+		if (ret < 0)
+			goto err_free_scan_elements;
+	}
 
 	ret = detect_and_move_global_attrs(dev);
-	if (ret < 0) {
-		free_device(dev);
-		return ret;
+	if (ret < 0)
+		goto err_free_device;
+
+	/* sorting is done after global attrs are added */
+	for (i = 0; i < dev->nb_channels; i++) {
+		struct iio_channel *chn = dev->channels[i];
+		qsort(chn->attrs,  chn->nb_attrs, sizeof(struct iio_channel_attr),
+			iio_channel_attr_compare);
 	}
+	qsort(dev->attrs, dev->nb_attrs, sizeof(char *),
+		iio_device_attr_compare);
 
 	dev->words = (dev->nb_channels + 31) / 32;
 	if (dev->words) {
 		mask = calloc(dev->words, sizeof(*mask));
 		if (!mask) {
-			free_device(dev);
-			return ret;
+			ret = -ENOMEM;
+			goto err_free_device;
 		}
 	}
 
 	dev->mask = mask;
 
 	ret = add_device_to_context(ctx, dev);
-	if (ret < 0)
-		free_device(dev);
+	if (!ret)
+		return 0;
+
+err_free_scan_elements:
+	for (i = 0; i < dev->nb_channels; i++)
+		free_protected_attrs(dev->channels[i]);
+err_free_device:
+	local_free_pdata(dev);
+	free_device(dev);
 	return ret;
 }
 
@@ -1481,7 +1810,7 @@ static int add_debug_attr(void *d, const char *path)
 {
 	struct iio_device *dev = d;
 	const char *attr = strrchr(path, '/') + 1;
-	char **attrs, *name = strdup(attr);
+	char **attrs, *name = iio_strdup(attr);
 	if (!name)
 		return -ENOMEM;
 
@@ -1557,42 +1886,11 @@ static const struct iio_backend_ops local_ops = {
 	.cancel = local_cancel,
 };
 
-static void init_index(struct iio_channel *chn)
-{
-	char buf[1024];
-	long id = -ENOENT;
-
-	if (chn->is_scan_element) {
-		id = (long) local_read_chn_attr(chn, "index", buf, sizeof(buf));
-		if (id > 0)
-			id = atol(buf);
-	}
-	chn->index = id;
-}
-
-static void init_data_format(struct iio_channel *chn)
+static void init_data_scale(struct iio_channel *chn)
 {
 	char buf[1024];
 	ssize_t ret;
 
-	if (chn->is_scan_element) {
-		ret = local_read_chn_attr(chn, "type", buf, sizeof(buf));
-		if (ret < 0) {
-			chn->format.length = 0;
-		} else {
-			char endian, sign;
-
-			sscanf(buf, "%ce:%c%u/%u>>%u", &endian, &sign,
-					&chn->format.bits, &chn->format.length,
-					&chn->format.shift);
-			chn->format.is_signed = (sign == 's' || sign == 'S');
-			chn->format.is_fully_defined =
-					(sign == 'S' || sign == 'U'||
-					chn->format.bits == chn->format.length);
-			chn->format.is_be = endian == 'b';
-		}
-	}
-
 	ret = iio_channel_attr_read(chn, "scale", buf, sizeof(buf));
 	if (ret < 0) {
 		chn->format.with_scale = false;
@@ -1609,14 +1907,67 @@ static void init_scan_elements(struct iio_context *ctx)
 	for (i = 0; i < ctx->nb_devices; i++) {
 		struct iio_device *dev = ctx->devices[i];
 
-		for (j = 0; j < dev->nb_channels; j++) {
-			struct iio_channel *chn = dev->channels[j];
-			init_index(chn);
-			init_data_format(chn);
-		}
+		for (j = 0; j < dev->nb_channels; j++)
+			init_data_scale(dev->channels[j]);
 	}
 }
 
+#ifdef WITH_LOCAL_CONFIG
+static int populate_context_attrs(struct iio_context *ctx, const char *file)
+{
+	struct INI *ini;
+	int ret;
+
+	ini = ini_open(file);
+	if (!ini) {
+		/* INI file not present -> not an error */
+		if (errno == ENOENT)
+			return 0;
+		else
+			return -errno;
+	}
+
+	while (true) {
+		const char *section;
+		size_t len;
+
+		ret = ini_next_section(ini, &section, &len);
+		if (ret <= 0)
+			goto out_close_ini;
+
+		if (!strncmp(section, "Context Attributes", len))
+			break;
+	}
+
+	do {
+		const char *key, *value;
+		char *new_key, *new_val;
+		size_t klen, vlen;
+
+		ret = ini_read_pair(ini, &key, &klen, &value, &vlen);
+		if (ret <= 0)
+			break;
+
+		/* Create a dup of the strings read from the INI, since they are
+		 * not NULL-terminated. */
+		new_key = strndup(key, klen);
+		new_val = strndup(value, vlen);
+
+		if (!new_key || !new_val)
+			ret = -ENOMEM;
+		else
+			ret = iio_context_add_attr(ctx, new_key, new_val);
+
+		free(new_key);
+		free(new_val);
+	} while (!ret);
+
+out_close_ini:
+	ini_close(ini);
+	return ret;
+}
+#endif
+
 struct iio_context * local_create_context(void)
 {
 	int ret = -ENOMEM;
@@ -1647,16 +1998,30 @@ struct iio_context * local_create_context(void)
 		goto err_set_errno;
 	}
 
-	snprintf(ctx->description, len + 5, "%s %s %s %s %s", uts.sysname,
+	iio_snprintf(ctx->description, len + 5, "%s %s %s %s %s", uts.sysname,
 			uts.nodename, uts.release, uts.version, uts.machine);
 
 	ret = foreach_in_dir(ctx, "/sys/bus/iio/devices", true, create_device);
 	if (ret < 0)
 		goto err_context_destroy;
 
+	qsort(ctx->devices, ctx->nb_devices, sizeof(struct iio_device *),
+		iio_device_compare);
+
 	foreach_in_dir(ctx, "/sys/kernel/debug/iio", true, add_debug);
 
 	init_scan_elements(ctx);
+
+#ifdef WITH_LOCAL_CONFIG
+	ret = populate_context_attrs(ctx, "/etc/libiio.ini");
+	if (ret < 0)
+		goto err_context_destroy;
+#endif
+
+	ret = iio_context_add_attr(ctx, "local,kernel", uts.release);
+	if (ret < 0)
+		goto err_context_destroy;
+
 	ret = iio_context_init(ctx);
 	if (ret < 0)
 		goto err_context_destroy;
@@ -1680,19 +2045,32 @@ int local_context_scan(struct iio_scan_result *scan_result)
 {
 	struct iio_context_info **info;
 	bool exists = false;
+	char *desc, *uri;
 	int ret;
 
-	ret = foreach_in_dir(&exists, "/sys/bus/iio/devices",
-			true, check_device);
+	ret = foreach_in_dir(&exists, "/sys/bus/iio", true, check_device);
 	if (ret < 0 || !exists)
 		return 0;
 
-	info = iio_scan_result_add(scan_result, 1);
-	if (!info)
+	desc = iio_strdup("Local devices");
+	if (!desc)
 		return -ENOMEM;
 
-	info[0]->description = "Local devices";
-	info[0]->uri = "local:";
+	uri = iio_strdup("local:");
+	if (!uri)
+		goto err_free_desc;
 
+	info = iio_scan_result_add(scan_result, 1);
+	if (!info)
+		goto err_free_uri;
+
+	info[0]->description = desc;
+	info[0]->uri = uri;
 	return 0;
+
+err_free_uri:
+	free(uri);
+err_free_desc:
+	free(desc);
+	return -ENOMEM;
 }
