@@ -37,6 +37,7 @@ static const struct option options[] = {
 	  {"buffer-size", required_argument, 0, 'b'},
 	  {"samples", required_argument, 0, 's' },
 	  {"timeout", required_argument, 0, 'T'},
+	  {"auto", no_argument, 0, 'a'},
 	  {0, 0, 0, 0},
 };
 
@@ -47,7 +48,8 @@ static const char *options_descriptions[] = {
 	"Use the specified trigger.",
 	"Size of the capture buffer. Default is 256.",
 	"Number of samples to capture, 0 = infinite. Default is 0.",
-	"Buffer timeout in milliseconds. 0 = no timeout"
+	"Buffer timeout in milliseconds. 0 = no timeout",
+	"Scan for available contexts and if only one is available use it.",
 };
 
 static void usage(void)
@@ -55,7 +57,7 @@ static void usage(void)
 	unsigned int i;
 
 	printf("Usage:\n\t" MY_NAME " [-n <hostname>] [-t <trigger>] "
-			"[-b <buffer-size>] [-s <samples>] "
+			"[-T <timeout-ms>] [-b <buffer-size>] [-s <samples>] "
 			"<iio_device> [<channel> ...]\n\nOptions:\n");
 	for (i = 0; options[i].name; i++)
 		printf("\t-%c, --%s\n\t\t\t%s\n",
@@ -81,7 +83,7 @@ static void quit_all(int sig)
 
 #ifdef _WIN32
 
-#include <Windows.h>
+#include <windows.h>
 
 BOOL WINAPI sig_handler_fn(DWORD dwCtrlType)
 {
@@ -97,7 +99,7 @@ BOOL WINAPI sig_handler_fn(DWORD dwCtrlType)
 	}
 }
 
-static setup_sig_handler(void)
+static void setup_sig_handler(void)
 {
 	SetConsoleCtrlHandler(sig_handler_fn, TRUE);
 }
@@ -140,12 +142,12 @@ static void setup_sig_handler(void)
 static void * sig_handler_thd(void *data)
 {
 	sigset_t *mask = data;
-	int ret;
+	int ret, sig;
 
 	/* Blocks until one of the termination signals is received */
 	do {
-		ret = sigwaitinfo(mask, NULL);
-	} while (ret == -1 && errno == EINTR);
+		ret = sigwait(mask, &sig);
+	} while (ret == EINTR);
 
 	quit_all(ret);
 
@@ -182,30 +184,6 @@ static void setup_sig_handler(void)
 
 #endif
 
-static struct iio_device * get_device(const struct iio_context *ctx,
-		const char *id)
-{
-
-	unsigned int i, nb_devices = iio_context_get_devices_count(ctx);
-	struct iio_device *device;
-
-	for (i = 0; i < nb_devices; i++) {
-		const char *name;
-		device = iio_context_get_device(ctx, i);
-		name = iio_device_get_name(device);
-		if (name && !strcmp(name, id))
-			break;
-		if (!strcmp(id, iio_device_get_id(device)))
-			break;
-	}
-
-	if (i < nb_devices)
-		return device;
-
-	fprintf(stderr, "Device %s not found\n", id);
-	return NULL;
-}
-
 static ssize_t print_sample(const struct iio_channel *chn,
 		void *buf, size_t len, void *d)
 {
@@ -220,51 +198,98 @@ static ssize_t print_sample(const struct iio_channel *chn,
 	return (ssize_t) len;
 }
 
+static struct iio_context *scan(void)
+{
+	struct iio_scan_context *scan_ctx;
+	struct iio_context_info **info;
+	struct iio_context *ctx = NULL;
+	unsigned int i;
+	ssize_t ret;
+
+	scan_ctx = iio_create_scan_context(NULL, 0);
+	if (!scan_ctx) {
+		fprintf(stderr, "Unable to create scan context\n");
+		return NULL;
+	}
+
+	ret = iio_scan_context_get_info_list(scan_ctx, &info);
+	if (ret < 0) {
+		char err_str[1024];
+		iio_strerror(-ret, err_str, sizeof(err_str));
+		fprintf(stderr, "Scanning for IIO contexts failed: %s\n", err_str);
+		goto err_free_ctx;
+	}
+
+	if (ret == 0) {
+		printf("No IIO context found.\n");
+		goto err_free_info_list;
+	}
+
+	if (ret == 1) {
+		ctx = iio_create_context_from_uri(iio_context_info_get_uri(info[0]));
+	} else {
+		fprintf(stderr, "Multiple contexts found. Please select one using --uri:\n");
+
+		for (i = 0; i < (size_t) ret; i++) {
+			fprintf(stderr, "\t%d: %s [%s]\n", i,
+				iio_context_info_get_description(info[i]),
+				iio_context_info_get_uri(info[i]));
+		}
+	}
+
+err_free_info_list:
+	iio_context_info_list_free(info);
+err_free_ctx:
+	iio_scan_context_destroy(scan_ctx);
+
+	return ctx;
+}
+
 int main(int argc, char **argv)
 {
 	unsigned int i, nb_channels;
 	unsigned int buffer_size = SAMPLES_PER_READ;
-	int c, option_index = 0, arg_index = 0, ip_index = 0, uri_index = 0;
+	const char *arg_uri = NULL;
+	const char *arg_ip = NULL;
+	int c, option_index = 0;
 	struct iio_device *dev;
 	size_t sample_size;
 	int timeout = -1;
+	bool scan_for_context = false;
 
-	while ((c = getopt_long(argc, argv, "+hn:u:t:b:s:T:",
+	while ((c = getopt_long(argc, argv, "+hn:u:t:b:s:T:a",
 					options, &option_index)) != -1) {
 		switch (c) {
 		case 'h':
 			usage();
 			return EXIT_SUCCESS;
 		case 'n':
-			arg_index += 2;
-			ip_index = arg_index;
+			arg_ip = optarg;
 			break;
 		case 'u':
-			arg_index += 2;
-			uri_index = arg_index;
+			arg_uri = optarg;
+			break;
+		case 'a':
+			scan_for_context = true;
 			break;
 		case 't':
-			arg_index += 2;
-			trigger_name = argv[arg_index];
+			trigger_name = optarg;
 			break;
 		case 'b':
-			arg_index += 2;
-			buffer_size = atoi(argv[arg_index]);
+			buffer_size = atoi(optarg);
 			break;
 		case 's':
-			arg_index += 2;
-			num_samples = atoi(argv[arg_index]);
+			num_samples = atoi(optarg);
 			break;
 		case 'T':
-			arg_index += 2;
-			timeout = atoi(argv[arg_index]);
+			timeout = atoi(optarg);
 			break;
 		case '?':
 			return EXIT_FAILURE;
 		}
 	}
 
-	if (arg_index + 1 >= argc) {
+	if (argc == optind) {
 		fprintf(stderr, "Incorrect number of arguments.\n\n");
 		usage();
 		return EXIT_FAILURE;
@@ -272,10 +297,12 @@ int main(int argc, char **argv)
 
 	setup_sig_handler();
 
-	if (uri_index)
-		ctx = iio_create_context_from_uri(argv[uri_index]);
-	else if (ip_index)
-		ctx = iio_create_network_context(argv[ip_index]);
+	if (scan_for_context)
+		ctx = scan();
+	else if (arg_uri)
+		ctx = iio_create_context_from_uri(arg_uri);
+	else if (arg_ip)
+		ctx = iio_create_network_context(arg_ip);
 	else
 		ctx = iio_create_default_context();
 
@@ -287,15 +314,18 @@ int main(int argc, char **argv)
 	if (timeout >= 0)
 		iio_context_set_timeout(ctx, timeout);
 
-	dev = get_device(ctx, argv[arg_index + 1]);
+	dev = iio_context_find_device(ctx, argv[optind]);
 	if (!dev) {
+		fprintf(stderr, "Device %s not found\n", argv[optind]);
 		iio_context_destroy(ctx);
 		return EXIT_FAILURE;
 	}
 
 	if (trigger_name) {
-		struct iio_device *trigger = get_device(ctx, trigger_name);
+		struct iio_device *trigger = iio_context_find_device(
+				ctx, trigger_name);
 		if (!trigger) {
+			fprintf(stderr, "Trigger %s not found\n", trigger_name);
 			iio_context_destroy(ctx);
 			return EXIT_FAILURE;
 		}
@@ -320,7 +350,7 @@ int main(int argc, char **argv)
 
 	nb_channels = iio_device_get_channels_count(dev);
 
-	if (argc == arg_index + 2) {
+	if (argc == optind + 1) {
 		/* Enable all channels */
 		for (i = 0; i < nb_channels; i++)
 			iio_channel_enable(iio_device_get_channel(dev, i));
@@ -328,7 +358,7 @@ int main(int argc, char **argv)
 		for (i = 0; i < nb_channels; i++) {
 			unsigned int j;
 			struct iio_channel *ch = iio_device_get_channel(dev, i);
-			for (j = arg_index + 2; j < (unsigned int) argc; j++) {
+			for (j = optind + 1; j < (unsigned int) argc; j++) {
 				const char *n = iio_channel_get_name(ch);
 				if (!strcmp(argv[j], iio_channel_get_id(ch)) ||
 						(n && !strcmp(n, argv[j])))
